#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/collector/T02ConfigCollector.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 28.10.2019
# Last Modified Date: 01.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import hashlib, json
from typing import Dict, Any, Optional, Union, List
from ampel.util.mappings import build_unsafe_short_dict_id
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class T02ConfigCollector(AbsDictConfigCollector):
	"""
	Collect and hash T2 configurations contained in T0 processes
	(hence the name T02)
	Note: used hash algorithm is blake2.
	An additional hash collision test is performed using SHA256
	"""

	def add(self,
		arg: Union[Dict[str, Any], List[str]],
		file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> int:

		if not isinstance(arg, dict):
			raise ValueError("Illegal argument")

		# Uses blake2 hash algorithm
		hh = build_unsafe_short_dict_id(arg)

		if self.get(hh):

			# Build SHA256 hash
			hash1 = hashlib.sha256(bytes(json.dumps(self.get(hh), sort_keys=True), "utf-8")).digest()
			hash2 = hashlib.sha256(bytes(json.dumps(arg, sort_keys=True), "utf-8")).digest()

			if hash1 != hash2:
				raise ValueError("Hash collision detected")

			if self.verbose:
				self.logger.log(VERBOSE, f"Re-using T2 config hash: {hh}")

			return hh

		if self.verbose:
			self.logger.log(VERBOSE, f"Adding T2 config hash: {hh}")

		self.__setitem__(hh, arg)
		return hh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/builder/ConfigChecker.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 03.09.2019
# Last Modified Date: 26.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>


import json
from typing import Any, Dict, Generator, Optional, Tuple
from ampel.log.AmpelLogger import AmpelLogger, DEBUG, ERROR
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.UnitLoader import UnitLoader
from ampel.dev.DictSecretProvider import PotemkinSecretProvider


class BaseConfigChecker:
	"""
	Validates a config usually build by ConfigBuilder
	"""

	def __init__(self, config: Dict[str, Any], logger: Optional[AmpelLogger] = None, verbose: bool = False):

		self.verbose = verbose
		self.logger = AmpelLogger.get_logger(
			console={'level': DEBUG if verbose else ERROR}
		) if logger is None else logger

		self.loader = UnitLoader(AmpelConfig(config), secrets=PotemkinSecretProvider())

		# Fast deep copy and serialization check
		self.config = json.loads(json.dumps(config))


	def iter_procs(self,
		ignore_inactive: bool = False,
		raise_exc: bool = False
	) -> Generator[Tuple[str, str], None, None]:

		for tier in ("t0", "t1", "t2", "t3"):

			procs = list(self.config['process'][tier].keys())

			for proc in procs:

				if ignore_inactive and not self.config['process'][tier][proc].get('active', True):
					self.logger.info("Ignoring inactivated processor model", extra={"process": proc})
					continue

				yield (tier, proc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsStateT2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 30.04.2020
# Last Modified Date: 08.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Union, Set, Tuple, List, TypeVar, Generic, Optional
from ampel.base import abstractmethod
from ampel.type import ChannelId
from ampel.ingest.compile.CompilerBase import CompilerBase
from ampel.ingest.CompoundBluePrint import CompoundBluePrint

T = TypeVar("T", bound=CompoundBluePrint)

class AbsStateT2Compiler(Generic[T], CompilerBase, abstract=True):

	@abstractmethod
	def compile(self,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
		compound_blueprint: T
	) -> Dict[Tuple[str, Optional[int], Union[bytes, Tuple[bytes, ...]]], Set[ChannelId]]:
		"""
		Build a set of T2 documents to be created, de-duplicating those that
		are requested by multiple channels via :func:`add_ingest_model`
	
		:param chan_selection: channels to create T2 documents for
		:param compound_blueprint: compound creation plan
		:returns:
		  A dict whose keys are
		    - T2 unit name
		    - id of T2 unit configuration
		    - compound id
		  and whose values are a set of channel ids
		"""
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/unit/base/DemoFirstPointT2Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2020
# Last Modified Date: 14.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from typing import Dict
from ampel.type import T2UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoFirstPointT2Unit(AbsPointT2Unit):

	ingest: Dict = {'eligible': 'first'}

	def run(self, datapoint: DataPoint) -> T2UnitResult:
		return {"time": time()}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/dev/DevAmpelContext.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.06.2020
# Last Modified Date: 10.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Any, Dict, Union, List
from ampel.core import AmpelContext, UnitLoader
from ampel.util.freeze import recursive_unfreeze
from ampel.config.AmpelConfig import AmpelConfig
from ampel.db.AmpelDB import AmpelDB
from ampel.model.ChannelModel import ChannelModel
from ampel.util.mappings import set_by_path, build_unsafe_short_dict_id
from ampel.dev.DictSecretProvider import PotemkinSecretProvider


class DevAmpelContext(AmpelContext):


	def __init__(self,
		db_prefix: Optional[str] = None, purge_db: bool = False,
		custom_conf: Optional[Dict[str, Any]] = None, **kwargs
	) -> None:
		"""
		:db_prefix: customizes the db prefix name
		(ex: "AmpelTest" will result in the databases: AmpelTest_data, AmpelTest_var, AmpelTest_ext)
		:purge_db: wheter all databases with the given prefix should be deleted (purged)
		:extra_conf: convenience parameter which allows to overwrite given parameters of the underlying
		ampel config (possibly frozen) dictionnary. Nested dict keys such as 'general.something' are supported.
		"""

		super().__init__(**kwargs)

		if db_prefix:
			dict.__setitem__(self.config._config['db'], 'prefix', db_prefix)

		if custom_conf or db_prefix:
			conf = self._get_unprotected_conf()
			for k, v in (custom_conf or {}).items():
				set_by_path(conf, k, v)
			self._set_new_conf(conf)

		if not self.loader.secrets:
			self.loader.secrets = PotemkinSecretProvider()

		if purge_db:
			self.db.drop_all_databases()
			self.db.init_db()


	def add_channel(self, name: Union[int, str], access: List[str] = []):
		cm = ChannelModel(channel=name, access=access)
		conf = self._get_unprotected_conf()
		for k, v in cm.dict().items():
			set_by_path(conf, f"channel.{name}.{k}", v)
		self._set_new_conf(conf)


	def add_config_id(self, arg: Dict[str, Any]) -> int:
		conf_id = build_unsafe_short_dict_id(arg)
		if conf_id in self.config.get("confid", dict, raise_exc=True):
			return conf_id
		conf = self._get_unprotected_conf()
		conf["confid"][conf_id] = arg
		self._set_new_conf(conf)
		return conf_id


	def _get_unprotected_conf(self) -> Dict[str, Any]:
		if self.config.is_frozen():
			return recursive_unfreeze(self.config._config) # type: ignore[arg-type]
		return self.config._config


	def _set_new_conf(self, conf: Dict[str, Any]) -> None:
		self.config = AmpelConfig(conf, True)
		self.loader = UnitLoader(self.config, secrets=self.loader.secrets)
		self.db = AmpelDB.new(self.config, self.loader.secrets)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/template/AbsLegacyChannelTemplate.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.10.2019
# Last Modified Date: 24.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from importlib import import_module
from pydantic import validator
from typing import List, Dict, Any, Union, Tuple, Optional, Sequence
from ampel.log.AmpelLogger import AmpelLogger
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.model.UnitModel import UnitModel
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate


class T2UnitModel(UnitModel):
	"""
	Precursor that ProcessMorpher can turn into a T2IngestModel
	"""
	#: Ingester options
	ingest: Optional[Dict[str, Any]]
	#: Filter result codes that should trigger this T2. If not specified, T2
	#: documents will be created in response to *any* passing alert.
	group: Union[int, List[int]] = []


class AbsLegacyChannelTemplate(AbsChannelTemplate, abstract=True):
	"""
	Abstract class whose purpose is to maintain compatibility with channel
	definitions created for ampel versions < 0.7.
	This class must be subclassed.
	
	Known subclass: :class:`~ampel.model.ZTFLegacyChannelTemplate.ZTFLegacyChannelTemplate`
	"""
	#: How to treat photopoints be treated once a transient has been accepted.
	#
	#: - false: apply filter to all photopoints
	#: - true or "live": bypass filter once a transient has been accepted once
	auto_complete: Union[bool, str]
	#: Filter to apply to incoming datapoints
	t0_filter: UnitModel
	#: T2 units to trigger when transient is updated. Dependencies of tied
	#: units will be added automatically.
	t2_compute: List[T2UnitModel] = []
	#: T3 processes bound to this channel. These may be use templates, such as
	#: :class:`~ampel.model.template.PeriodicSummaryT3.PeriodicSummaryT3`.
	t3_supervise: List[Dict[str, Any]] = []


	@validator('t3_supervise', 't2_compute', pre=True, each_item=False)
	def cast_to_list_if_required(cls, v):
		if isinstance(v, dict):
			return [v]
		return v


	@validator('auto_complete')
	def make_auto_stock_match(cls, v):
		if v is True or v == 'live':
			return {
				'filter': 'bypass',
				'update_rej': True,
				'retro_complete': True
			}
		else:
			return {
				'filter': 'overrule',
				'update_rej': False,
				'retro_complete': False
			}


	# Mandatory implementation
	def get_channel(self, logger: AmpelLogger) -> Dict[str, Any]:
		d = self.dict(by_alias=True)
		for k in ("auto_complete", "t0_filter", "t2_compute", "t3_supervise"):
			if k in d:
				del d[k]
		return d


	def craft_t0_process(self,
		first_pass_config: FirstPassConfig,
		controller: Dict[str, Any],
		stock_ingester: Union[str, Tuple[str, Dict[str, Any]]],
		t0_ingester: Union[str, Tuple[str, Dict[str, Any]]],
		t1_ingester: Optional[Union[str, Tuple[str, Dict[str, Any]]]] = None,
		t1_standalone_ingester: Optional[Union[str, Tuple[str, Dict[str, Any]]]] = None,
		t2_state_ingester: Optional[Union[str, Tuple[str, Dict[str, Any]]]] = None,
		t2_point_ingester: Optional[Union[str, Tuple[str, Dict[str, Any]]]] = None,
		t2_stock_ingester: Optional[Union[str, Tuple[str, Dict[str, Any]]]] = None,
		t2_compute_from_t0: List[T2UnitModel] = [],
		t2_compute_from_t1: List[T2UnitModel] = [],
	) -> Dict[str, Any]:
		"""
		This method needs a reference to a FirstPassConfig dict because
		config information might be needed during the template transforming process.
		For example, legacy channel templates (such as ZTFLegacyChannelTemplate)
		allow users to reference any kind of t2 units under the root config section 't2_compute'.
		The AlertProcessor however, requires different configuration paths for "state T2s" and "point T2s".
		The underlying templates will thus have to sort T2s based on their respective abstract classes,
		and for this, the ampel configuration is required.

		:param stock_ingester: unit_class or (unit_class, config dict)
		:param t0_ingester: unit_class or (unit_class, config dict)
		:param t1_ingester: unit_class or (unit_class, config dict)
		:param t2_state_ingester: unit_class or (unit_class, config dict)
		:param t2_point_ingester: unit_class or (unit_class, config dict)
		:param t2_stock_ingester: unit_class or (unit_class, config dict)
		:param t2_compute_from_t0: units to schedule on t0_add
		:param t2_compute_from_t1: units to schedule on t1_combine
		"""

		ret: Dict[str, Any] = {
			"tier": 0,
			"schedule": ["super"],
			"active": self.active,
			"distrib": self.distrib,
			"source": self.source,
			"channel": self.channel,
			"name": f"{self.channel}|T0|{self.template}",
			"controller": controller,
			"processor": {
				"unit": "AlertProcessor",
				"config": {
					"directives": [{
						"channel": self.channel,
						"stock_match": self.auto_complete,
						"filter": self.t0_filter.dict(exclude_unset=True, by_alias=True),
						"t0_add": self._get_dict(t0_ingester),
						"stock_update": self._get_dict(stock_ingester)
					}]
				}
			}
		}

		directives = ret['processor']['config']['directives'][0]

		self.check_tied_units(
			t2_compute_from_t0 + t2_compute_from_t1,
			first_pass_config
		)

		if t2_state_units := self.get_units(
			t2_compute_from_t0,
			[
				"AbsStateT2Unit",
				"AbsCustomStateT2Unit",
				"AbsTiedStateT2Unit",
				"AbsTiedCustomStateT2Unit",
			],
			first_pass_config
		):

			if t1_ingester is None:
				raise ValueError("Template processing requires parameter 't1_ingester'")

			if t2_state_ingester is None:
				raise ValueError("Template processing requires parameter 't2_state_ingester'")

			directives['t0_add']['t1_combine'] = [
				self._get_dict(
					t1_ingester,
					t2_compute = self._get_dict(
						t2_state_ingester,
						units = t2_state_units
					)
				)
			]

		if t2_state_units := self.get_units(
			t2_compute_from_t1,
			[
				"AbsStateT2Unit",
				"AbsCustomStateT2Unit",
				"AbsTiedStateT2Unit",
				"AbsTiedCustomStateT2Unit",
			],
			first_pass_config
		):

			if t1_standalone_ingester is None:
				raise ValueError("Template processing requires parameter 't1_standalone_ingester'")

			if t2_state_ingester is None:
				raise ValueError("Template processing requires parameter 't2_state_ingester'")

			directives['t1_combine'] = [
				self._get_dict(
					t1_standalone_ingester,
					t2_compute = self._get_dict(
						t2_state_ingester,
						units = t2_state_units
					)
				)
			]

		if self.get_units(t2_compute_from_t1, ["AbsStockT2Unit", "AbsPointT2Unit"], first_pass_config):
			raise NotImplementedError("Stock and Point T2 units can't be scheduled from standalone T1 processes")

		if t2_stock_units := self.get_units(t2_compute_from_t0, "AbsStockT2Unit", first_pass_config):

			if t2_stock_ingester is None:
				raise ValueError("Template processing requires parameter 't2_stock_ingester'")

			directives['t2_compute'] = self._get_dict(
				t2_stock_ingester, units = t2_stock_units
			)

		if t2_point_units := self.get_units(t2_compute_from_t0, "AbsPointT2Unit", first_pass_config):

			if t2_point_ingester is None:
				raise ValueError("Template processing requires parameter 't2_point_ingester'")

			directives['t0_add']['t2_compute'] = self._get_dict(
				t2_point_ingester, units = t2_point_units
			)

		return ret


	def _get_dict(self, arg: Union[str, Tuple[str, Dict[str, Any]]], **kwargs) -> Dict[str, Any]:
		""" internal use """
		if isinstance(arg, tuple):
			if arg[1]:
				return {"unit": arg[0], "config": arg[1], **kwargs}
			return {"unit": arg[0], **kwargs}
		return {"unit": arg, **kwargs}


	def check_tied_units(self,
		all_t2_units: List[T2UnitModel],
		first_pass_config: FirstPassConfig
	) -> None:
		"""
		:raises: ValueError if tied t2 units are present in t2_units but the requred t2 units are not present in t2_compute
		"""
		tied_units: List[T2UnitModel] = []
		for el in all_t2_units:
			if "AbsTiedT2Unit" in first_pass_config['unit']['base'][el.unit_name]['base']:
				tied_units.append(el) # type: ignore
			
		s = set()
		for tied_unit in tied_units:
			klass = getattr(
				import_module(
					first_pass_config['unit']['base'][tied_unit.unit_name]['fqn']
				),
				tied_unit.unit_name
			)
			s.update(klass.get_tied_unit_names())

		if missing_deps := s.difference({u.unit for u in all_t2_units}):
			raise ValueError(
				f"Following t2 unit(s) must be defined (required by "
				f"tied units): {list(missing_deps)}"
			)
		
		def as_unitmodel(t2_unit_model: T2UnitModel) -> UnitModel:
			return UnitModel(**{k: v for k,v in t2_unit_model.dict().items() if k in UnitModel.__fields__})

		for tied_unit in tied_units:
			if (
				isinstance(tied_unit.config, dict)
				and (t2_deps := (tied_unit.config.get("t2_dependency") or []))
			):
				for t2_dep in t2_deps:
					dependency_config = UnitModel(unit=t2_dep["unit"], config=t2_dep.get("config"))
					candidates = [as_unitmodel(unit) for unit in all_t2_units if unit.unit_name == dependency_config.unit_name]
					if not any((c == dependency_config for c in candidates)):
						raise ValueError(
							f"Unit {tied_unit.unit_name} depends on unit {dependency_config.dict()}, "
							f"which was not configured. Possible matches are: "
							f"{[c.dict() for c in candidates]}"
						)


	def get_units(self,
		units: Sequence[UnitModel], abs_unit: Union[str, List[str]],
		first_pass_config: FirstPassConfig
	) -> List[Dict]:
		"""
		:returns: unit defintions (dict) that are subclass of the provided abstract class name.
		"""

		if isinstance(abs_unit, str):
			abs_unit = [abs_unit]

		return [
			el.dict(exclude_unset=True, by_alias=True)
			for el in units
			if any(
				unit in first_pass_config['unit']['base'][el.unit]['base']
				for unit in abs_unit
			)
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/__init__.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.06.2020
# Last Modified Date: 16.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# flake8: noqa: F401
from .AmpelContext import AmpelContext
from .AdminUnit import AdminUnit
from .UnitLoader import UnitLoader
from .JournalUpdater import JournalUpdater



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/filter/AbsT3Filter.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 12.02.2020
# Last Modified Date: 20.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence
from ampel.base import AmpelABC, AmpelBaseModel, abstractmethod
from ampel.core.AmpelBuffer import AmpelBuffer


class AbsT3Filter(AmpelBaseModel, AmpelABC, abstract=True):

	@abstractmethod
	def filter(self, abufs: Sequence[AmpelBuffer]) -> Sequence[AmpelBuffer]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/demo/unit/base/DemoPointT2Unit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 25.03.2020
# Last Modified Date: 14.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from ampel.type import T2UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoPointT2Unit(AbsPointT2Unit):

	test_parameter: int = 1

	def run(self, datapoint: DataPoint) -> T2UnitResult:
		return {
			"datapoint_id": datapoint['_id'],
			"time": time(),
			"test_parameter": self.test_parameter
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/log/LogsBufferDict.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 02.04.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import sys
from typing import List, Dict, Any
if sys.version_info.minor > 8:
	from typing import TypedDict
else:
	from typing_extensions import TypedDict


class LogsBufferDict(TypedDict, total=False):
	"""
	Allows to concatenate various log entries with different 'extra' parameters
	(that in this case normally do not self-aggregate automatically).
	This class is used for example by the AlertProcessor to ensure the aggregation
	of log entries emitted by different ingesters (T0, T1, T2)
	"""
	logs: List[str]
	extra: Dict[str, Any]
	err: bool



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/StockIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.12.2017
# Last Modified Date: 09.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from pymongo import UpdateOne
from typing import Tuple, Dict, List, Any, Union, Literal, Optional
from ampel.type import StockId, ChannelId
from ampel.abstract.ingest.AbsStockIngester import AbsStockIngester
from ampel.util.mappings import flatten_dict


class StockIngester(AbsStockIngester):

	# setOnInsert tag
	tag: Optional[List[Union[int, str]]]
	tier: Literal[0, 1, 3] = 0


	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.set_on_insert = {'tag': self.tag}


	def ingest(self,
		stock_id: StockId,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]],
		jextra: Dict[str, Any]
	) -> None:
		"""
		Update transient document
		:param jextra: includes provided parameter into the journal extra's field
		"""

		now = int(time())
		created = {'any': now}
		modified = {'any': now}
		jchan: List[ChannelId] = []

		# loop through all channels,
		for chan, res in chan_selection:
			# json requires dict keys to be str
			if isinstance(chan, int):
				c = str(chan)
				created[c] = now
				modified[c] = now
			else:
				created[chan] = now
				modified[chan] = now
			jchan.append(chan)

		if len(jchan) == 1:
			jchan = jchan[0] # type: ignore[assignment]
			chan_add_to_set = jchan
		else:
			chan_add_to_set = {'$each': jchan} # type: ignore[assignment]

		# Insert/Update transient document into stock collection
		self.updates_buffer.add_stock_update(
			UpdateOne(
				{'_id': stock_id},
				{
					'$addToSet': {'channel': chan_add_to_set},
					'$setOnInsert': self.get_setOnInsert(stock_id),
					'$min': flatten_dict({'created': created}),
					'$set': flatten_dict({'modified': modified}),
					'$push': {
						'journal': {
							'tier': self.tier,
							'ts': now,
							'channel': jchan,
							'run': self.run_id,
							'extra': jextra
						}
					}
				},
				upsert=True
			)
		)


	def get_setOnInsert(self, stock_id: StockId) -> Dict[str, Any]:
		""" To be overriden by sub-classes """
		return self.set_on_insert



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/utils/general.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.04.2020
# Last Modified Date: 10.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Type, Literal, Union
from ampel.type import strict_iterable, StrictIterable


def has_nested_type(obj: StrictIterable, target_type: Type, strict: bool = True) -> bool:
	"""
	:param obj: object instance (dict/list/set/tuple)
	:param target_type: example: ReadOnlyDict/list
	:param strict: must be an instance of the provided type (subclass instances would be rejected)
	"""

	if strict:
		# pylint: disable=unidiomatic-typecheck
		if type(obj) is target_type:
			return True
	else:
		if isinstance(obj, target_type):
			return True

	if isinstance(obj, dict):
		for el in obj.values():
			if has_nested_type(el, target_type):
				return True

	elif isinstance(obj, strict_iterable):
		for el in obj:
			if has_nested_type(el, target_type):
				return True

	return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/context/T3AddAlertsNumber.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.01.2020
# Last Modified Date: 20.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, ClassVar
from ampel.db.query.events import build_t0_stats_query
from ampel.t3.context.AbsT3RunContextAppender import AbsT3RunContextAppender
from ampel.t3.context.T3AddLastRunTime import T3AddLastRunTime


class T3AddAlertsNumber(AbsT3RunContextAppender):

	key: ClassVar[str] = "processed_alerts"

	def update(self, run_context: Dict[str, Any]) -> None:
		""" Add number of alerts processed since last run of this event as "processed_alerts" """

		if T3AddLastRunTime.key not in run_context:
			T3AddLastRunTime(
				context=self.context,
				logger=self.logger,
				process_name=self.process_name
			).update(run_context) # updates context dict

		if not run_context[T3AddLastRunTime.key]:
			self.logger.info(
				"Last run time not available, cannot count number of alerts"
			)
			run_context[self.key] = None
			return

		# Get number of alerts processed since last run
		res = next(
			self.context.db.get_collection('events').aggregate(
				build_t0_stats_query(gte_time=run_context[T3AddLastRunTime.key])
			),
			None
		)

		run_context[self.key] = 0 if res is None else res.get('alerts')



from typing import Dict, Any, Optional

from ampel.core.AdminUnit import AdminUnit
from ampel.base import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger


class AbsOpsUnit(AdminUnit, abstract=True):
    """
    A unit for performing sceduled maintenance tasks not associated with a
    particular processing tier: collecting metrics, reporting exceptions, etc.
    """

    logger: AmpelLogger

    @abstractmethod
    def run(self, beacon: Optional[Dict[str,Any]]=None) -> Optional[Dict[str,Any]]:
        """
        :param beacon: the result of the previous run
        :returns:
          a BSON-serializable document summarizing the run. This will be
          supplied to the next invocation as `beacon`.
        """
        ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/context/AbsT3RunContextAppender.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.01.2020
# Last Modified Date: 10.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any
from ampel.base import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger
from ampel.core.AdminUnit import AdminUnit


class AbsT3RunContextAppender(AdminUnit, abstract=True):
	"""
	Add contextual information to a T3 process, such as the time of the previous
	run, or the number of alerts processed since that run.
	
	Inherits from AdminUnit because subclasses might need access to
	the AmpelConfig (foremost to the contained resource definitions)
	"""

	logger: AmpelLogger
	process_name: str

	@abstractmethod
	def update(self, run_context: Dict[str, Any]) -> None:
		"""
		:param run_context: run context to fill
		"""
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/builder/ConfigBuilder.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 03.09.2019
# Last Modified Date: 01.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import importlib, re, json
from math import inf # noqa: required for eval(repr(...)) below
from typing import Dict, List, Any, Optional, Set, Iterable

from pydantic import ValidationError

from ampel.util.mappings import get_by_path, set_by_path
from ampel.util.crypto import aes_recursive_decrypt
from ampel.abstract.AbsChannelTemplate import AbsChannelTemplate
from ampel.log.AmpelLogger import AmpelLogger, VERBOSE, DEBUG, ERROR
from ampel.config.builder.FirstPassConfig import FirstPassConfig
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.T02ConfigCollector import T02ConfigCollector
from ampel.model.template.ChannelWithProcsTemplate import ChannelWithProcsTemplate
from ampel.config.collector.ProcessConfigCollector import ProcessConfigCollector
from ampel.config.collector.ChannelConfigCollector import ChannelConfigCollector
from ampel.config.builder.ProcessMorpher import ProcessMorpher

class ConfigBuilder:
	"""
	Builds a central configuration dict for ampel. Config building is a two pass process:
	First, all available configuration files are loaded from different repositories
	and information merged together (into the instance self.first_pass_config)
	Then, the config is 'morphed' into its final structure by a multi-step process (method build_config)
	"""

	_default_processes = ["DefaultT2Process", "DefaultPurge"]

	def __init__(self, logger: AmpelLogger = None, verbose: bool = False):

		self.logger = AmpelLogger.get_logger(console={'level': DEBUG if verbose else ERROR}) if logger is None else logger
		self.first_pass_config = FirstPassConfig(logger, verbose)
		self.templates: Dict[str, Any] = {}
		self.verbose = verbose
		self.error = False


	def load_ampel_conf(self,
		d: Dict, file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> None:

		if self.verbose:
			self.logger.log(VERBOSE, f"Loading global ampel conf ({file_name}) from repo {dist_name}")

		# "db" "logging" "channel" "unit" "process" "alias" "resource"
		for k in self.first_pass_config.conf_keys:

			if k not in d:
				continue

			if k in ('unit', 'process', 'alias'):
				if isinstance(d[k], list):
					self.first_pass_config[k].add(d[k], file_name=file_name, dist_name=dist_name)
				elif isinstance(d[k], dict):
					# kk = 'processor', 'base', ... for root key "unit"
					# kk = 't0', 't1', 't2', 't3', ... for root key "process" or "alias"
					for kk, v in d[k].items(): # kk = 'processor', 'base', ... for root key "unit"
						if kk in self.first_pass_config[k]:
							if self.verbose:
								self.logger.log(VERBOSE, f"Parsing {k}.{kk}")
							self.first_pass_config[k][kk].add(v, file_name=file_name, dist_name=dist_name)
						else:
							self.logger.error(f"Unknown config element: {k}.{kk}")

			else:
				self.first_pass_config[k].add(d[k], file_name=file_name, dist_name=dist_name)

		if 'template' in d:
			self.register_channel_templates(d['template'], file_name, dist_name)


	'''
	def load_conf_section(self,
		section: str, arg: Dict, file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> None:
		"""
		Depending on the value of parameter 'section', a structure may be expected for dict 'arg'.
		1) tier-less sections: 'channel', 'db', 'resource'
		-> no structure imposed
		2) tier-dependent sub-sections: 'controller', 'processor', 'unit', 'alias', 'process'
		-> arg must have the following JSON structure:
		{'t0': { ... }, 't1': { ... }, ...}
		whereby the t0, t1, t2 and t3 keys are optional (at least one is required though)
		"""

		# ('channel', 'db', 'resource')
		if section in self.first_pass_config.general_keys:
			self.first_pass_config[section].add(
				arg, file_name=file_name, dist_name=dist_name
			)
			return

		# ('controller', 'processor', 'unit', 'alias', 'process')
		if section in self.first_pass_config.tier_keys:
			for k in ('t0', 't1', 't2', 't3'):
				if k in arg:
					self.first_pass_config[k][section].add(
						arg[k], file_name=file_name, dist_name=dist_name
					)

		raise ValueError(f'Unknown config section: {section}')
	'''


	def register_channel_templates(self,
		chan_templates: Dict[str, str],
		file_name: Optional[str] = None,
		dist_name: Optional[str] = None
	) -> None:

		if not isinstance(chan_templates, dict):
			raise ValueError('Provided argument must be a dict instance')

		for k, v in chan_templates.items():

			if k in self.templates:
				raise ValueError('Duplicated channel template: ' + k)

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Registering template "{k}" ' +
					file_name if file_name else '' +
					ConfigCollector.distrib_hint(distrib=dist_name)
				)

			self.templates[k] = getattr(
				importlib.import_module(v),
				v.split('.')[-1]
			)


	def build_config(self,
		stop_on_errors: int = 2,
		config_validator: Optional[str] = "ConfigChecker",
		skip_default_processes: bool = False,
		pwds: Optional[Iterable[str]] = None
	) -> Dict[str, Any]:
		"""
		Pass 2.
		Builds the final ampel config using previously collected config pieces (contained in self.first_pass_config)
		This involves a multi-step process where the config is 'morphed' into its final structure.

		:param stop_on_errors: by default, config building stops and raises an exception if an error occured.
			- 2: stop on errors
			- 1: ignore errors in first_pass_config only (will stop on morphing/scoping/template errors)
			- 0: ignore all errors
		Note that issues might/will later arise with your ampel system.

		:param pwds: config section 'resource' might contain AES encrypted entries.
		If passwords are provided to this method, thoses entries will be decrypted.

		:param skip_default_processes: set to True to discard default processes defined by ampel-core.
		The static variable ConfigBuilder._default_processes references those processes by name.
		Set skip_default_processes=True if your repositories define their own default T2/T3 processes.

		:raises: ValueError if self.error is True - this behavior can be disabled using the parameter stop_on_errors
		"""

		if self.first_pass_config.has_nested_error():
			if stop_on_errors > 1:
				raise ValueError(
					'Error were reported in first pass config, you can use the option stop_on_errors = 1 (or 0)\n' +
					'to bypass this exception and get the (possibly non-working) config nonetheless'
				)

		out = {
			k: self.first_pass_config[k]
			for k in FirstPassConfig.conf_keys.keys()
		}

		out['process'] = {}

		# Add t2 init config collector (in which both hashed values of t2 run configs
		# and t2 init config will be added)
		out['confid'] = T02ConfigCollector(
			conf_section='confid', logger=self.logger, verbose=self.verbose
		)

		# Add (possibly transformed) processes to output config
		for tier in (0, 1, 2, 3, "ops"):

			tier_name = tier if tier == "ops" else f"t{tier}"

			if self.verbose:
				self.logger.log(VERBOSE, f'Checking standalone {tier_name} processes')

			p_collector = ProcessConfigCollector(
				tier=tier, conf_section='process', # type: ignore[arg-type]
				logger=self.logger, verbose=self.verbose
			)

			#out['process'][f't{tier}'] = {
			#	k: self.first_pass_config[f't{tier}'][k]
			#	for k in FirstPassConfig.tier_keys if k != 'process'
			#}

			# New empty process collector to gather morphed processes
			out['process'][tier_name] = p_collector

			# For each process collected before, apply transformations
			# and add it to our (almost) final process collector.
			# 'almost' because a gathering of T0 processes may occure later
			for p in self.first_pass_config['process'][tier_name].values():

				if skip_default_processes and p["name"] in self._default_processes:
					self.logger.info(f'Skipping {tier_name} default processes: {p["name"]}')
					continue

				if self.verbose:
					self.logger.log(VERBOSE, f'Morphing standalone {tier_name} processes: {p["name"]}')
					pass

				try:
					p_collector.add(
						self.new_morpher(p) \
							.scope_aliases(self.first_pass_config) \
							.apply_template() \
							.hash_t2_config(out) \
							.get(),
						p.get('source'),
						p.get('distrib')
					)
				except Exception as e:
					self.logger.error(f'Unable to morph process {p["name"]}', exc_info=e)
					if stop_on_errors > 0:
						raise e

		# Setup empty channel collector
		out['channel'] = ChannelConfigCollector(
			conf_section='channel', logger=self.logger, verbose=self.verbose
		)

		# Fill it with (possibly transformed) channels
		for chan_name, chan_dict in self.first_pass_config['channel'].items():

			# Template processing is required for this particular channel
			if tpl := self._get_channel_tpl(chan_dict):

				# Extract channel definition from template instance
				try:
					out['channel'].add(
						tpl.get_channel(self.logger)
					)
				except (ValidationError, Exception) as ee:
					if isinstance(ee, ValidationError):
						self.logger.error(f'Unable to morph channel: {chan_name}')
						self.logger.error(str(ee))
					else:
						self.logger.error(f'Unable to morph channel: {chan_name}', exc_info=ee)
					if stop_on_errors > 0:
						raise ee

				# Retrieve processes possibly embedded in channel def
				for p in tpl.get_processes(self.logger, self.first_pass_config):

					if self.verbose:
						self.logger.log(VERBOSE,
							f'Morphing channel embedded t{p["tier"]} process: {p["name"]}'
						)

					try:
						# Add transformed process to final process collector
						out['process'][f't{p["tier"]}'].add(
							self.new_morpher(p) \
								.scope_aliases(self.first_pass_config) \
								.apply_template() \
								.hash_t2_config(out) \
								.enforce_t3_channel_selection(chan_name) \
								.get(),
							p.get('source'),
							p.get('distrib')
						)
					except (ValidationError, Exception) as ee:
						if isinstance(ee, ValidationError):
							self.logger.error(f'Unable to morph embedded process {p["name"]} (from {p["source"]})')
							self.logger.error(str(ee))
						else:
							self.logger.error(f'Unable to morph embedded process {p["name"]} (from {p["source"]})', exc_info=ee)
						if stop_on_errors > 0:
							raise ee

			else:

				# Raw/Simple/Standard channel definition
				# (encouraged behavior actually)
				out['channel'].add(chan_dict)


		# Optionaly decrypt encrypted config entries
		if pwds:
			out['resource'] = aes_recursive_decrypt(out['resource'], pwds)

		self.logger.info('Done building config')

		# Casts ConfigCollector instances into real dicts
		d = self._recursive_dictify(out)

		if config_validator:
			from importlib import import_module
			validator = getattr(
				import_module("ampel.config.builder." + config_validator),
				config_validator
			)(d, self.logger, self.verbose)
			return validator.validate()

		return d


	@classmethod
	def _recursive_dictify(cls, item):
		if isinstance(item, dict):
			return {k: cls._recursive_dictify(v) for k, v in item.items()}
		elif isinstance(item, list):
			return [cls._recursive_dictify(v) for v in item]
		else:
			return item


	def new_morpher(self, process: Dict[str, Any]) -> ProcessMorpher:
		"""
		Returns an instance of ProcessMorpher using the provided
		process dict and the internal logger and templates
		"""
		return ProcessMorpher(
			process, self.templates, self.logger, self.verbose
		)


	def _get_channel_tpl(self, chan_dict: Dict[str, Any]) -> Optional[AbsChannelTemplate]:
		"""
		Internal method used to check if a template (and which one)
		should be applied to a given channel dict.
		"""

		if 'template' in chan_dict:

			if self.verbose:
				self.logger.log(VERBOSE,
					f'Channel {chan_dict["channel"]} ({chan_dict["source"]}) '
					f'declares template: {chan_dict["template"]}'
				)

			if chan_dict['template'] not in self.templates:
				raise ValueError(f'Unknown template: {chan_dict["template"]}')

			return self.templates[chan_dict['template']](**chan_dict)


		if 'process' in chan_dict:
			return ChannelWithProcsTemplate(**chan_dict)

		return None


	def gather_processes(self,
		config: FirstPassConfig, tier: int, match: str, collect: str
	) -> Optional[Dict]:
		"""
		:param channel_names:
		- None: all the available channels from the ampel config will be loaded
		- String: channel with the provided id will be loaded
		- List of strings: channels with the provided ids will be loaded
		"""

		processes: List[Dict] = [
			el for el in config[f't{tier}']['process'].values()
			if re.match(match, el.get('name'))
		]

		if len(processes) <= 1:
			return None

		init_configs = []
		dist_names: Set[str] = set()
		out_proc: Optional[Dict] = None

		for p in processes:

			if not p.get('active', False):
				self.logger.info(f'Ignoring deactivated process: {p["name"]}')
				continue

			# Use first active proces as template for the multi-channel process
			# to be generated by this method
			if not out_proc:
				# Faster deep copy
				out_proc = json.loads(json.dumps(p))

			# Inactivate this particular process as it will be run in a multi-channel process
			self.logger.info(f'Deactivating process: {p["name"]}')
			p['active'] = False

			# Gather init config
			init_configs.append(
				get_by_path(p, collect)
			)

			# Collect distribution name
			if p.get('distrib'):
				dist_names.add(p.get('distrib')) # type: ignore

		if out_proc is None:
			return None

		# for T0 processes: collect=processor.config.directives
		set_by_path(out_proc, collect, init_configs)
		out_proc['name'] = match
		out_proc['distrib'] = '/'.join(dist_names)

		return out_proc


	def print(self) -> None:
		self.first_pass_config.print()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/complement/AbsT3DataAppender.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 16.01.2020
# Last Modified Date: 14.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Iterable
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.base import abstractmethod
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.core.AdminUnit import AdminUnit

# Inherits AdminUnit because implementing classes might need access to
# an AmpelConfig instance (foremost to the contained resource definitions)
class AbsT3DataAppender(AdminUnit, abstract=True):
	"""
	Complement :class:`~ampel.core.AmpelBuffer.AmpelBuffer` with information
	stored outside the Ampel database.
	"""

	logger: LoggerProtocol

	@abstractmethod
	def complement(self, it: Iterable[AmpelBuffer]) -> None:
		"""Add information to each :class:`~ampel.core.AmpelBuffer.AmpelBuffer` """
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : ampel/view/plot/SVGBrowser.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.06.2019
# Last Modified Date: 15.06.2019
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# type: ignore[import]

from ampel.db.AmpelDB import AmpelDB
from ampel.plot.SVGCollection import SVGCollection
from ampel.plot.SVGLoader import SVGLoader
from IPython.display import HTML, display
from collections import defaultdict
import random

class SVGBrowser:
	"""
	"""

	def __init__(self, svg_loader, scale=None, show_col_title=True, show_svg_titles=True):
		""" """

		self._scale = scale
		self._svg_loader = svg_loader
		self._show_col_title = show_col_title
		self._show_svg_titles = show_svg_titles


	def show_multi_window(self):
		""" 
		Shows all plots for one transient in one window
		"""
		pass


	def show_single_window(
		self, scale=None, append_tran_name=True, 
		win_title="Ampel plots", tran_ids=None, 
		png_convert=False, multiproc=0, 
		global_flex_box=False
	):
		""" 
		:param bool append_tran_name: appends transient name to plot titles
		"""

		win_id, dom_id = SVGBrowser._new_window(win_title)

		if global_flex_box:
			dom_id = SVGBrowser._insert_flex_box(win_id)

		if multiproc:

			from concurrent.futures import ProcessPoolExecutor
			futures = []

			with ProcessPoolExecutor(max_workers=multiproc) as executor:

				for tran_id in self._svg_loader._plots.keys():
					futures.append(
						executor.submit(
							get_html, 
							tran_id, 
							self._svg_loader._data_query,
							self._svg_loader._t2_query,
							scale,
							png_convert,
							not global_flex_box
						)
					)

				for future in futures:
					SVGBrowser._write_to_dom(
						dom_id, future.result(), 
						'Adding %s info to %s<br>' % (tran_id, win_title)
					)
		else:

			for tran_id in self._svg_loader._plots:

				if tran_ids:
					if self._svg_loader._plots[tran_id] not in tran_ids:
						continue

				SVGBrowser._write_to_dom(
					dom_id, 
					self._svg_loader._plots[tran_id]._repr_html_(
						scale=scale, 
						title_prefix=tran_id, 
						png_convert=png_convert,
						flexbox_wrap=not global_flex_box
					), 
					'Adding %s info to %s<br>' % (tran_id, win_title)
				)


		display(HTML("Done"))


	@staticmethod
	def _new_window(title):
		"""
		:returns: js variable *name* pointing to body DOM element within the newly created window
		"""

		rand_str = str(random.randint(0, 100000))
		win_id = "win_" + rand_str
		body_id = "body_" + rand_str

		display(
			HTML(
				'<script type="text/Javascript"> \
					var %s=window.open("", "%s"); \
					%s.document.title="%s"; \
					var %s=%s.document.body; \
				 </script>' % (win_id, title, win_id, title, body_id, win_id)
			)
		)

		return win_id, body_id


	@staticmethod
	def _write_to_dom(js_body_var_name, html_content, feedback=""):
		""" """

		display(
			HTML(
				'%s<script type="text/Javascript"> \
					%s.innerHTML += \'%s\';' \
					#%s.document.body.innerHTML += \'%s\';' \
				'</script>' % (
					feedback, 
					js_body_var_name, 
					html_content.replace("\n","\\")
				)
			)
		)


	@staticmethod
	def _insert_flex_box(js_body_var_name):
		"""
		:returns: js variable *name* pointing to (flexbox)
		div DOM element within the newly created window
		"""
		flex_id = "flex_" + str(random.randint(0, 100000))

		display(
			HTML(
				'<script type="text/Javascript"> \
					%s.document.body.innerHTML += \'<div id=%s style="\
						text-align:center; \
						display: flex; \
						flex-direction: row; \
						flex-wrap: wrap; \
						justify-content: center"></div>\'; \
					var %s = %s.document.getElementById("%s");\
				</script>' % (
					js_body_var_name, flex_id, 
					flex_id, js_body_var_name, flex_id
				)
			)
		)

		return flex_id


def get_html(
	tran_id, data_query, t2_query, scale=1.0, 
	png_convert=False, flexbox_wrap_svg_col=True
):
	""" 
	"""

	svg_loader = SVGLoader()
	svg_loader._data_query = data_query
	svg_loader._t2_query = t2_query
	svg_loader.load_plots()

	return svg_loader._plots[tran_id]._repr_html_(
		scale=scale, 
		title_prefix=tran_id, 
		png_convert=png_convert,
		flexbox_wrap=flexbox_wrap_svg_col
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/LogicSchemaUtils.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.10.2018
# Last Modified Date: 16.03.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.type import strict_iterable
from ampel.util.collections import check_seq_inner_type
from ampel.config.LogicSchemaIterator import LogicSchemaIterator
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf

class LogicSchemaUtils:


	@staticmethod
	def iter(arg):
		"""
		Please see :obj:`LogicSchemaIterator <ampel.config.t3.LogicSchemaIterator>` \
		docstring for more info
		"""
		return LogicSchemaIterator(arg)


	@staticmethod
	def to_logical_struct(v, field_name):

		if isinstance(v, list):
			raise ValueError(
				"transients->select->%s config error\n" % field_name +
				"'%s' parameter cannot be a list. " % field_name +
				"Please use the following syntax:\n" +
				" -> {'any_of': ['Ab', 'Cd']} or\n" +
				" -> {'all_of': ['Ab', 'Cd']} or\n" +
				" -> {'one_of': ['Ab', 'Cd']} or\n" +
				"One level nesting is allowed, please see\n" +
				"ConfigUtils.conditional_expr_converter(..) docstring for more info"
			)

		if isinstance(v, (str, int)):
			return {'any_of': [v]}

		if isinstance(v, dict):

			if len(v) != 1:
				raise ValueError(
					"transients->select->%s config error\n" % field_name +
					"Unsupported dict format %s" % v
				)

			if 'any_of' in v:

				if not isinstance(v['any_of'], strict_iterable):
					raise ValueError(
						"transients->select->%s:any_of config error\n" % field_name +
						"Invalid dict value type: %s. Must be a sequence" % type(v['any_of'])
					)

				# 'any_of' supports only a list of dicts and str/int
				if not check_seq_inner_type(v['any_of'], (str, int, dict), multi_type=True):
					raise ValueError(
						"transients->select->%s:any_of config error\n" % field_name +
						"Unsupported nesting (err 2)"
					)

				if not check_seq_inner_type(v['any_of'], (int, str)) and len(v['any_of']) < 2:
					raise ValueError(
						"transients->select->%s:any_of config error\n" % field_name +
						"any_of list must contain more than one element when containing all_of\n" +
						"Offending value: %s" % v
					)

				for el in v['any_of']:

					if isinstance(el, dict):

						if 'any_of' in el:
							raise ValueError(
								"transients->select->%s:any_of.any_of config error\n" % field_name +
								"Unsupported nesting (any_of in any_of)"
							)

						if 'all_of' in el:

							# 'all_of' closes nesting
							if not check_seq_inner_type(el['all_of'], (int, str)):
								raise ValueError(
									"transients->select->%s:any_of.all_of config error\n" % field_name +
									"Unsupported nesting (all_of list content must be str/int)"
								)

							if len(set(el['all_of'])) < 2:
								raise ValueError(
									"transients->select->%s:all_of config error\n" % field_name +
									"Please do not use all_of with just one element\n" +
									"Offending value: %s" % el
								)

						else:
							raise ValueError(
								"transients->select->%s:any_of config error\n" % field_name +
								"Unsupported nested dict: %s" % el
							)

			elif 'all_of' in v:

				# 'all_of' closes nesting
				if (
					not isinstance(v['all_of'], strict_iterable) or
					not check_seq_inner_type(v['all_of'], (int, str))
				):
					raise ValueError(
						"transients->select->%s:all_of config error\n" % field_name +
						"Invalid type for value %s\n(must be a sequence, is: %s)\n" %
						(v['all_of'], type(v['all_of'])) +
						"Note: no nesting is allowed below 'all_of'"
					)

				if len(set(v['all_of'])) < 2:
					raise ValueError(
						"transients->select->%s:all_of config error\n" % field_name +
						"Please do not use all_of with just one element\n" +
						"Offending value: %s" % v
					)

			elif 'one_of' in v:

				# 'one_of' closes nesting
				if (
					not isinstance(v['one_of'], strict_iterable) or
					not check_seq_inner_type(v['one_of'], (int, str))
				):
					raise ValueError(
						"transients->select->%s:one_of config error\n" % field_name +
						"Invalid type for value %s\n(must be a sequence, is: %s)\n" %
						(v['one_of'], type(v['one_of'])) +
						"Note: no nesting is allowed below 'one_of'"
					)

			else:
				raise ValueError(
					"transients->select->%s config error\n" % field_name +
					"Invalid dict key (only 'any_of', 'all_of', 'one_of' are allowed)"
				)

		return v


	@staticmethod
	def reduce_to_set(arg, in_type=(str, int)):
		"""
		.. sourcecode:: python\n
			for schema in (a,b,c,d,e):
				print("Schema: %s" % schema)
				print("Reduced set: %s" % LogicSchemaUtils.reduce_to_set(schema))

			Schema: 'a'
			Reduced set: {'a'}
			Schema: {'any_of': ['a', 'b', 'c']}
			Reduced set: {'b', 'a', 'c'}
			Schema: {'all_of': ['a', 'b', 'c']}
			Reduced set: {'b', 'a', 'c'}
			Schema: {'any_of': [{'all_of': ['a', 'b']}, 'c']}
			Reduced set: {'b', 'a', 'c'}
			Schema: {'any_of': [{'all_of': ['a', 'b']}, {'all_of': ['a', 'c']}, 'd']}
			Reduced set: {'d', 'b', 'a', 'c'}
		"""

		if isinstance(arg, in_type):
			return {arg}

		if isinstance(arg, (AllOf, AnyOf, OneOf)):
			arg = arg.dict()

		if isinstance(arg, dict):

			if "any_of" in arg:
				s = set()
				for el in arg['any_of']:
					if isinstance(el, in_type):
						s.add(el)
					elif isinstance(el, dict):
						for ell in next(iter(el.values())):
							s.add(ell)
					else:
						raise ValueError("LogicSchemaUtils.reduce_to_set: unsupported format (1)")
				return s

			if 'all_of' in arg:
				return set(arg['all_of'])

			if 'one_of' in arg:
				return set(arg['one_of'])

			raise ValueError("LogicSchemaUtils.reduce_to_set: unsupported format (2)")

		raise ValueError("LogicSchemaUtils.reduce_to_set: unsupported type: %s" % type(arg))


	@classmethod
	def conditional_expr_converter(cls, arg, level=1):
		"""
		Converts JSON encoded conditional statements from Ampel config file
		into arrays with dimension up to two.
		'any_of' -> or operator -> encoded in a array elements of depth=1
		'all_of' -> and operator -> encoded in array elements of depth=2

		Accepted input:
		---------------

		atomar values str, int float: "a" / 1 / 1.2

		1d sequences of atomar values (automaticall_ofy treated as 'any_of' sequence):
		[1, 2, 3]  / [1, "a", 3.4]

		'any_of' dict containing 1d list of atomar values (explicit 'any_of' sequence):
		{'any_of': [1, 2, 3]} / {'any_of': [1, "a", 3.4]}

		'all_of' dict containing 1d list of atomar values
		{'all_of': [1, 2, 3]} / {'any_of': [1, "a", 3.4]}

		Nested structure whereby 'all_of' closes the nesting (can contain only a sequence of atomar values)
		{
			'any_of': [
				{'all_of': ["HUSN1", "HUSN2"]},
				"HUBH1",
				{'all_of': ["HUSN1", "HUSN3"]}
			]
		}

		Examples:
		---------

		```
		In []: conditional_expr_converter("abc")
		Out[]: 'abc'

		In []: conditional_expr_converter(["1","2","3"])
		Out[]: ["1", "2", "3"]

		In []: conditional_expr_converter({'all_of': ["3", "1", "2"]})
		Out[]: [["3", "1", "2"]]

		In []: conditional_expr_converter({'any_of': ["3", "1", "2"]})
		Out[]: ["3", "1", "2"]

		In []: conditional_expr_converter({'any_of': [{'all_of': ["1","2"]}, "3", "1", "2"]})
		Out[]: [["1", "2"], "3", "1", "2"]

		In []: conditional_expr_converter({'any_of': [{'all_of': ["1","2"]}, "3", {'all_of': ["1","3"]}]})
		Out[]: [["1", "2"], "3", ["1", "3"]]

		In []: conditional_expr_converter(["1", "2", ["1", "2""3"]])
		ValueError: Unsupported format (0)

		In []: conditional_expr_converter({'all_of': ["1", "2", ["1","2"]]})
		ValueError: Unsupported nesting

		In []: conditional_expr_converter({'all_of': ["1", "2"], 'abc': "2"})
		ValueError: Unsupported format ("1")

		In []: conditional_expr_converter({'any_of': [{'any_of': ["1","2"]}, "2"]})
		ValueError: Unsupported nesting

		In []: conditional_expr_converter({'any_of': [{'all_of': ["1","2"]}, "3", {'any_of': ["1","2"]}]})
		ValueError: Unsupported nesting

		In []: conditional_expr_converter({'all_of': [{'all_of': ["1","2"]}, "3", "1", "2"]})
		ValueError: Unsupported nesting

		In []: conditional_expr_converter({'any_of': [{'all_of': ["1","2"]}, "3", {'all_of': ["1",{'all_of':["1","2"]}]}]})
		ValueError: Unsupported nesting
		```
		"""

		sequences = (list, tuple)
		ok = (str, int, float)

		if level > 2:
			raise ValueError("Unsupported nesting level")

		if isinstance(arg, (str, int, float)):
			return arg

		if isinstance(arg, sequences):
			if not check_seq_inner_type(arg, ok, multi_type=True):
				raise ValueError("Unsupported format (0)")
			return arg

		if isinstance(arg, dict):

			if len(arg) != 1:
				raise ValueError("Unsupported format (1)")

			key = next(iter(arg.keys()))

			if key == "all_of":

				# Value must be a sequence
				if not isinstance(arg[key], sequences):
					raise ValueError("Unsupported format (3)")

				# 'all_of' closes nesting (content must be atomar elements of type 'ok')
				if not check_seq_inner_type(arg[key], ok, multi_type=True):
					raise ValueError("Unsupported nesting")

				return [arg[key]] if level == 1 else arg[key]

			if key == "any_of":

				if level > 1:
					raise ValueError("Unsupported nesting")

				# Value must be a sequence
				if not isinstance(arg[key], sequences):
					raise ValueError("Unsupported format (4)")

				if check_seq_inner_type(arg[key], ok, multi_type=True):
					return arg[key]

				return [cls.conditional_expr_converter(el, level=level + 1) for el in arg[key]]

			raise ValueError("Unsupported format (5)")

		raise ValueError("Unsupported format (6)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsProcessorUnit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 04.02.2020
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Dict, Optional
from ampel.type import ChannelId
from ampel.base import abstractmethod
from ampel.core.AdminUnit import AdminUnit
from ampel.log.LogFlag import LogFlag


class AbsProcessorUnit(AdminUnit, abstract=True):
	"""
	Base class for units that orchestrate an Ampel processing tier.
	"""

	#: name of the process that instantiated this processor
	process_name: str
	#: channels associated with this processor
	channel: Optional[ChannelId] = None

	#: raise exceptions instead of catching and logging
	raise_exc: bool = False
	#: logging configuration to use, from the logging section of the Ampel config
	log_profile: str = "default"
	#: flag to apply to all log records created
	base_log_flag: LogFlag = LogFlag.SCHEDULED_RUN
	#: additional kwargs to pass to :class:`~ampel.log.handlers.DBLoggingHandler.DBLoggingHandler`
	db_handler_kwargs: Optional[Dict[str, Any]] = None


	@abstractmethod(var_args=True)
	def run(self) -> Any:
		"""
		Run processing steps.
		"""
		...


	def new_run_id(self) -> int:
		"""
		Return an identifier that can be used to associate log entries from a
		single process invocation. This ID is unique and monotonicaly increasing.
		"""
		return self.context.db \
			.get_collection('counter') \
			.find_one_and_update(
				{'_id': 'current_run_id'},
				{'$inc': {'value': 1}},
				new=True, upsert=True
			) \
			.get('value')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/T3Processor.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.02.2018
# Last Modified Date: 21.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Sequence, List, Optional, Union

from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
from ampel.log import AmpelLogger, LogFlag, DBEventDoc, SHOUT
from ampel.log.utils import report_exception
from ampel.model.t3.T3Directive import T3Directive
from ampel.t3.load.AbsT3Loader import AbsT3Loader
from ampel.t3.run.AbsT3UnitRunner import AbsT3UnitRunner
from ampel.t3.select.AbsT3Selector import AbsT3Selector
from ampel.t3.complement.AbsT3DataAppender import AbsT3DataAppender
from ampel.t3.context.AbsT3RunContextAppender import AbsT3RunContextAppender
from ampel.util.collections import chunks


class T3Processor(AbsProcessorUnit):
	"""
	:param update_journal: Record the invocation of this event in the stock journal
	:param update_events: Record this event in the events collection
	:param extra_journal_tag:
	  tag(s) to add to the :class:`~ampel.content.JournalRecord.JournalRecord`
	  of each selected stock

	.. note:: by default, :func:`run` add entries to the journals of selected
	  stocks, store log records in the database, and records the invocation in
	  the events collection. To run an "anonymous" event that does not change
	  the state of the database:
	
	  - use a log_profile without db logging (profiles are defined under 'logging' in the ampel config)
	  - set update_journal = False (no update to the transient doc)
	  - set update_events = False (no update to the events collection)
	  - set raise_exc = True (troubles collection will not be populated if an exception occurs)
	"""

	#: T3 processing sequences
	directives: Sequence[T3Directive]
	#: number of stocks to load at once
	chunk_size: int = 200


	def __init__(self,
		update_journal: bool = True,
		update_events: bool = True,
		extra_journal_tag: Optional[Union[int, str]] = None,
		**kwargs
	) -> None:
		"""
		Note that update_journal, update_event and extra_journal_tag are admin run options and
		should be set only on command line. They are thus not defined as part of the underlying model.
		"""

		super().__init__(**kwargs)

		self.update_journal = update_journal
		self.update_events = update_events
		self.extra_journal_tag = extra_journal_tag

		if 'db' not in self.context.config.get(f"logging.{self.log_profile}", dict, raise_exc=True):
			for el in ("update_journal", "update_events"):
				if getattr(self, el):
					raise ValueError(
						f"{el} cannot be True without a logger associated with a db logging handler"
					)


	def run(self) -> None:

		event_doc = None
		run_id = self.new_run_id()
		exc = None
		logger = None

		try:

			logger = AmpelLogger.from_profile(
				self.context, self.log_profile, run_id,
				base_flag = LogFlag.T3 | LogFlag.CORE | self.base_log_flag,
				force_refresh = True
			)

			# Admins have the option to to run a T3 process silently/anonymously
			if self.update_events:

				if not self.process_name:
					raise ValueError("Parameter process_name must be defined")

				# Create event doc
				event_doc = DBEventDoc(
					self.context.db, tier=3, run_id=run_id,
					process_name=self.process_name,
				)

			# Feedback
			logger.log(SHOUT, f"Running {self.process_name}")

			for directive in self.directives:

				# run context
				#############

				run_context: Dict[str, Any] = {}

				if self.context.admin_msg:
					run_context['admin_msg'] = self.context.admin_msg

				if directive.context:

					for el in directive.context:
						self.context.loader \
							.new_admin_unit(
								unit_model = el,
								context = self.context,
								sub_type = AbsT3RunContextAppender,
								logger = logger,
								process_name = self.process_name
							) \
							.update(run_context)

				# Unit runner
				#############

				# The default runner provided by pyampel-core is T3UnitRunner
				runner = self.context.loader \
					.new_admin_unit(
						unit_model = directive.run,
						context = self.context,
						sub_type = AbsT3UnitRunner,
						logger = logger,
						run_id = run_id,
						process_name = self.process_name,
						channel = self.channel,
						update_journal = self.update_journal,
						extra_journal_tag = self.extra_journal_tag,
						run_context = run_context
					)

				# target selection
				##################

				if directive.select:

					# Spawn and run a new selector instance
					# stock_ids is an iterable (often a pymongo cursor)
					selector = self.context.loader.new_admin_unit(
						unit_model = directive.select,
						context = self.context,
						sub_type = AbsT3Selector,
						logger = logger
					)

					# Usually, id_key is '_id' but it can be 'stock' if the
					# selection is based on t2 documents for example
					id_key = selector.field_name


					# Content loader
					################

					if directive.load:
						# Spawn requested content loader
						content_loader = self.context.loader \
							.new_admin_unit(
								unit_model = directive.load,
								context = self.context,
								sub_type = AbsT3Loader,
								logger = logger
							)


					# Content complementer
					######################

					# Spawn potentialy requested snapdata complementers
					if directive.complement:
						comps: List[AbsT3DataAppender] = [
							self.context.loader \
								.new_admin_unit(
									unit_model = conf_el,
									context = self.context,
									sub_type = AbsT3DataAppender,
									logger = logger
								)
							for conf_el in directive.complement
						]

					# get pymongo cursor
					if stock_ids := selector.fetch():

						# Run start
						###########

						# Loop over chunks from the cursor/iterator
						# NB: we consume the entire stock selection cursor at once using list()
						# to be robust against cursor timeouts or server restarts during long-
						# lived T3 processes
						for chunk_ids in chunks(list(stock_ids), self.chunk_size):

							# try/catch here to allow some chunks to complete
							# even if one raises an exception
							try:
								# Load info from DB
								tran_data = content_loader.load([sid[id_key] for sid in chunk_ids])

								# Potentialy add complementary information (spectra, TNS names, ...)
								if directive.complement:
									for appender in comps:
										appender.complement(tran_data)

								# Run T3 units defined for this process
								runner.run(list(tran_data))
							except Exception as e:
								exc = e
								if self.raise_exc:
									raise e

								report_exception(
									self.context.db, logger, exc=e,
									info={'process': self.process_name}
								)

				runner.done()

		except Exception as e:

			exc = e
			if self.raise_exc:
				raise e

			if not logger:
				logger = AmpelLogger.get_logger()

			report_exception(
				self.context.db, logger, exc=e,
				info={'process': self.process_name}
			)

		finally:

			if not logger:
				logger = AmpelLogger.get_logger()

			# Feedback
			logger.log(SHOUT, f"Done running {self.process_name}")
			logger.flush()

			# Register the execution of this event into the events col
			if event_doc:
				event_doc.update(logger, success=(exc is None))



import io
import subprocess
from argparse import Namespace

import pytest
import yaml

from ampel.abstract.AbsProcessorUnit import AbsProcessorUnit
from ampel.base.BadConfig import BadConfig
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.config.builder.ConfigChecker import ConfigChecker
from ampel.config.builder.ConfigValidator import ConfigValidator
from ampel.core.UnitLoader import UnitLoader
from ampel.config import cli


def test_build_config():
    config = yaml.safe_load(
        io.BytesIO(subprocess.check_output(["ampel-config", "build"]))
    )
    assert ConfigValidator(config).validate() == config


def test_ConfigChecker(testing_config, monkeypatch):
    """
    ConfigValidator validates units without calling their __init__ methods
    """
    with open(testing_config) as f:
        config = yaml.safe_load(f)
    # validates as-is
    checker = ConfigChecker(config)
    checker.validate(raise_exc=True)

    # add a processor with side-effects
    class SideEffect(RuntimeError):
        ...

    class SideEffectLadenProcessor(AbsProcessorUnit):

        required: int

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            raise SideEffect

        def run(self):
            ...

    config["process"]["t0"]["BadProcess"] = {
        "name": "BadProcess",
        "active": True,
        "schedule": "super",
        "processor": {"unit": "SideEffectLadenProcessor", "config": {"required": 1}},
    }
    checker = ConfigChecker(config)

    def get_class_by_name(name, *args, **kwargs):
        if name == "SideEffectLadenProcessor":
            return SideEffectLadenProcessor
        else:
            return UnitLoader.get_class_by_name(checker.loader, name, *args, **kwargs)

    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)

    # ConfigChecker attempts to instantiate the unit
    with pytest.raises(SideEffect):
        assert checker.validate(raise_exc=True)

    # ConfigValidator just validates the model
    checker = ConfigValidator(config)
    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)
    checker.validate(raise_exc=True)

    # ConfigValidator fails if the config does not satisfy the model
    checker.config["process"]["t0"]["BadProcess"]["processor"]["config"].clear()
    with pytest.raises(BadConfig):
        checker.validate()


@pytest.mark.parametrize("doc", [{"bignumber": 1 << 57}, {1: 2}])
def test_transform_config(doc):
    """Transform preserves objects that are not representable in JSON"""
    args = Namespace(
        config_file=io.StringIO(), filter=".", output_file=io.StringIO(), validate=False
    )
    yaml.dump(doc, args.config_file)
    args.config_file.seek(0)
    cli.transform(args)
    args.output_file.seek(0)
    transformed_doc = yaml.safe_load(args.output_file)
    assert transformed_doc == doc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/operator/FlatAnyOf.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.10.2018
# Last Modified Date: 15.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import collections
from typing import List, Generic
from pydantic import validator
from pydantic.generics import GenericModel
from ampel.type import T

class FlatAnyOf(GenericModel, Generic[T]):
	"""
	Similar to AnyOf except that it does not allow embedded AllOf elements
	"""
	any_of: List[T]

	@validator('any_of', pre=True)
	def cast_to_list(cls, v):
		if not isinstance(v, collections.abc.Sequence):
			return [v]
		return v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/JournalUpdater.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.10.2018
# Last Modified Date: 08.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from time import time
from bson import ObjectId
from pymongo.errors import BulkWriteError
from pymongo.operations import UpdateMany, UpdateOne
from typing import List, Any, Optional, Sequence, Union, Dict, Literal, get_args

from ampel.db.AmpelDB import AmpelDB
from ampel.type import ChannelId, Tag, StockId
from ampel.log import AmpelLogger, VERBOSE
from ampel.log.utils import report_exception
from ampel.struct.JournalTweak import JournalTweak
from ampel.content.JournalRecord import JournalRecord

tag_type = get_args(Tag) # type: ignore[misc]
chan_type = get_args(ChannelId) # type: ignore[misc]

class JournalUpdater:

	def __init__(self,
		ampel_db: AmpelDB, tier: Literal[0, 1, 2, 3], run_id: int,
		process_name: str, logger: AmpelLogger, raise_exc: bool = False,
		update_journal: bool = True,
		extra_tag: Optional[Union[Tag, Sequence[Tag]]] = None
	) -> None:
		"""
		:param raise_exc: raise exception rather than populating 'troubles' collection
		"""

		self._ampel_db = ampel_db
		self.col_stock = ampel_db.get_collection('stock')
		self.run_id = run_id
		self.tier = tier
		self.raise_exc = raise_exc
		self.update_journal = update_journal
		self.process_name = process_name
		self.extra_tag = extra_tag
		self.logger = logger
		self.reset()


	def reset(self) -> None:
		self.journal_updates: List[Any] = []
		self.journal_updates_count = 0


	def new_record(self,
		unit: Optional[Union[int, str]] = None,
		channels: Optional[Union[ChannelId, Sequence[ChannelId]]] = None,
		doc_id: Optional[ObjectId] = None,
		now: Optional[Union[int, float]] = None
	) -> JournalRecord:

		ret: JournalRecord = {
			'tier': self.tier,
			'ts': now if now else int(time()),
			'process': self.process_name,
			'run': self.run_id
		}

		if channels:
			ret['channel'] = channels

		if unit:
			ret['unit'] = unit

		if doc_id:
			ret['doc'] = doc_id.binary

		if self.extra_tag:
			ret['tag'] = self.extra_tag

		return ret


	def add_record(self,
		stock: Union[StockId, Sequence[StockId]],
		jextra: Optional[JournalTweak] = None,
		doc_id: Optional[ObjectId] = None,
		unit: Optional[Union[int, str]] = None,
		channel: Optional[Union[ChannelId, Sequence[ChannelId]]] = None,
		now: Optional[Union[int, float]] = None
	) -> JournalRecord:
		"""
		:returns: the JournalRecord dict instance associated with the stock document(s) update
		is returned by this method so that customization can be made if necessary.
		Note that the associated update operation does more than just adding a journal record,
		it also modifies the "modified" field of stock document(s).
		"""

		if isinstance(stock, Sequence):
			self.journal_updates_count += len(stock)
			match: Any = {'$in': list(stock)}
			Op = UpdateMany
		else:
			self.journal_updates_count += 1
			match = stock
			Op = UpdateOne

		jrec = self.new_record(unit, channel, doc_id, now)

		if jextra:

			if jextra.tag:
				self.include_tags(jrec, jextra.tag)

			if jextra.extra:
				jrec['extra'] = jextra.extra

			if jextra.status and jextra.status > 0:
				jrec['status'] = jextra.status

		if channel:
			if isinstance(channel, chan_type):
				maxd = {f'modified.{channel}': jrec['ts']}
			else:
				maxd = {
					f'modified.{chan}': jrec['ts']
					for chan in channel # type: ignore[union-attr]
				}
		else:
			maxd = {}

		maxd['modified.any'] = jrec['ts']

		if self.update_journal:
			self.journal_updates.append(
				Op(
					{'_id': match},
					{'$push': {'journal': jrec}, '$max': maxd}
				)
			)

		return jrec


	def flush(self) -> None:

		if not self.journal_updates:
			return

		jupds = self.journal_updates
		self.reset()

		try:

			if self.logger.verbose > 1:
				for el in jupds:
					self.logger.log(VERBOSE, f"Journal update: {str(el)}")

			self.col_stock.bulk_write(jupds)

			if self.logger.verbose:
				self.logger.log(VERBOSE,
					f"{len(jupds)} journal entr{'ies' if len(jupds) > 1 else 'y'} inserted"
				)

		except Exception as e:

			if self.raise_exc:
				if isinstance(e, BulkWriteError):
					raise ValueError(f"Journal update error: {e.details}")
				raise ValueError("Journal update error")

			info: Dict[str, Any] = {
				'process': self.process_name,
				'msg': 'Exception in flush()',
				'journalUpdateCount': self.journal_updates_count
			}

			if isinstance(e, BulkWriteError):
				info['BulkWriteError'] = str(e.details)

			# Populate troubles collection
			report_exception(self._ampel_db, self.logger, exc=e, info=info)


	@staticmethod
	def include_tags(jrec: JournalRecord, tag: Union[Tag, Sequence[Tag]]):
		""" Modifies the input JournalRecord dict """

		if tag:

			# journal record also contains tag(s)
			if 'tag' in jrec:

				# tweak request is about a single tag (not a sequence of tags)
				if isinstance(tag, tag_type):
					# journal record contains single tag (not a sequence of tags)
					if isinstance(jrec['tag'], tag_type):
						jrec['tag'] = [jrec['tag'], tag] # type: ignore[list-item]
					else:
						jrec['tag'].append(tag) # type: ignore[union-attr]

				# multi-tag tweak request
				else:
					# journal record contains single tag (not a sequence of tags)
					if isinstance(jrec['tag'], tag_type):
						jrec['tag'] = [*jrec['tag'], tag] # type: ignore[list-item, misc]
					else:
						jrec['tag'] = jrec['tag'] + tag # type: ignore[operator]

			# journal record contains no tag
			else:
				jrec['tag'] = tag




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/AdminUnit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.10.2019
# Last Modified Date: 15.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.core.AmpelContext import AmpelContext


class AdminUnit(AmpelABC, AmpelBaseModel, abstract=True):
	"""
	Top level abstract class receiving a reference to
	an AmpelContext instance as constructor parameter
	"""

	def __init__(self, context: AmpelContext, **kwargs):
		#: Configuration, database client, etc.
		self.context: AmpelContext = context
		AmpelBaseModel.__init__(self, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/StateT2Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.05.2020
# Last Modified Date: 01.05.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.abstract.ingest.AbsStateT2Ingester import AbsStateT2Ingester

class StateT2Ingester(AbsStateT2Ingester):
	""" To be implemented """
	pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/plot/SVGCollection.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.06.2019
# Last Modified Date: 09.02.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

# type: ignore[import]

from typing import Dict, Any, Optional, List
from ampel.plot.SVGUtils import SVGUtils
from ampel.plot.SVGPlot import SVGPlot

class SVGCollection:

	def __init__(self,
		title: str = None, scale: float = 1.0,
		inter_padding: int = 100, center: bool = True
	):
		"""
		:param title: title of this collection
		:param scale: scale factor for all SVGs (default: 1.0)
		:param inter_padding: sets padding in px between plots of this collection
		"""
		self._svgs = []
		self._col_title = title
		self._scale = scale
		self._inter_padding = inter_padding
		self._center = center


	def rescale(self, scale: float = 1.0) -> None:
		"""
		:param float scale: scale factor for all SVGs (default: 1.0)
		"""
		if self._scale == scale:
			return
		for el in self._svgs:
			el._content = SVGUtils.rescale(el._content, scale)
		self._scale = scale


	def set_inter_padding(self, inter_padding: int) -> None:
		"""
		Sets padding in px between plots of this collection
		"""
		self._inter_padding = inter_padding


	def add_svg_plot(self, svgp: SVGPlot) -> None:

		if not isinstance(svgp, SVGPlot):
			raise ValueError("Instance of ampel.plot.SVGPlot expected")

		self._svgs.append(svgp)


	def add_svg_dict(self, svg_dict: Dict[str, Any], title_left_padding: int = 0) -> None:
		"""
		:param Dict svg_dict:
		:param int title_left_padding:
		"""
		self._svgs.append(
			SVGPlot(
				content = svg_dict['svg'],
				title = svg_dict['title'],
				tags = svg_dict['tag'],
				title_left_padding = title_left_padding
			)
		)


	def add_raw_db_dict(self, svg_dict: Dict[str, Any]) -> None:
		"""
		:param Dict svg_dict: raw svg dict loaded from DB
		"""
		self.add_svg_dict(
			SVGUtils.decompress_svg_dict(svg_dict)
		)


	def get_svgs(self, tag: Optional[str] = None, tags: Optional[List[str]] = None) -> List[SVGPlot]:

		if tag:
			return [svg for svg in self._svgs if svg.has_tag(tag)]
		if tags:
			return [svg for svg in self._svgs if svg.has_tags(tag)]

		return self._svgs


	def _repr_html_(
		self, scale: Optional[float] = None, show_col_title: bool = True,
		title_prefix: Optional[str] = None, show_svg_titles: bool = True,
		hide_if_empty: bool = True, png_convert: bool = False,
		flexbox_wrap: bool = True
	) -> str:
		"""
		:param scale: if None, native scaling is used
		"""

		if hide_if_empty and not self._svgs:
			return None

		html = "<center>" if self._center else ""
		#html += '<hr style="width:100%; border: 2px solid;"/>'

		if show_col_title and self._col_title:
			html += '<h1 style="color: darkred">' + self._col_title + '</h1>'

		if flexbox_wrap:
			html += '<div style="\
				text-align:center; \
				display: flex; \
				flex-direction: row; \
				flex-wrap: wrap; \
				justify-content: center">'

		for svg in self._svgs:
			html += svg._repr_html_(
				scale = scale,
				show_title = show_svg_titles,
				title_prefix = title_prefix,
				padding_bottom = self._inter_padding,
				png_convert = png_convert
			)

		if flexbox_wrap:
			return html + "</div></center>" if self._center else html + "</div>"
		else:
			return html + "</center>" if self._center else html


	def show_html(self, **kwargs):
		"""
		:param **kwargs: see _repr_html_ arguments for details
		"""
		from IPython.display import HTML
		return HTML(
			self._repr_html_(**kwargs)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/db/LogsLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 29.11.2018
# Last Modified Date: 15.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.log.AmpelLogger import AmpelLogger
from ampel.view.ReadOnlyDict import ReadOnlyDict
from ampel.db.AmpelDB import AmpelDB
from ampel.log.LogFlag import LogFlag

# TODO: update this class
class LogsLoader:
	"""
	"""

	def __init__(self, read_only=True, logger=None):
		"""
		"""
		self.logger = AmpelLogger.get_logger() if logger is None else logger
		self.read_only = read_only


	def fetch_main_logs(
		self, matcher, decompactify=True, simplify=False, hexify=True, resolve_flag=True,
		remove=["channel"], datetime='string', datetime_key='_id', verbose=False
	):
		"""
		:param matcher: insance of LogsMatcher
		:type matcher: LogsMatcher
		:param bool decompactify:
		:param bool resolve_flag: \
			prints 'flag': <LogFlag.SCHEDULED_RUN|CORE|T2|INFO: 8836> \
			instead of 'flag': 8836
		:param bool simplify:
		:param datetime: if not None, datetime of each log entry (extracted from ObjectId) \
			will be added to the log content (use param datetime_key to configure whished key). \
			Possible values: \
			'string': datetime will be added as string: '2019-03-07T08:32:17.000Z' \
			'date': datetime will be added as datetime object: datetime.datetime(2019, 3, 7, 8, 32, 30)
		:param str datetime_key: \
			'_id': garanteed in first position but overrides default _id (don't use remove=["_id"]...) \
			or any string ('date' for example)
		:param bool verbose: whether to print the generated aggregation stages
		:param List[str] remove: possible values: 'channel', '_id', 'tranId'
		"""

		stages = [
			# Matching criteria (can contain nested dicts in case of complex criteria)
			{'$match': matcher.get_match_criteria()}
		]

		# Extract datetime from objectid and add it as 'date' field
		if datetime:
			stages.append(
				{
					'$addFields': {
						datetime_key: {
							"$convert": {
								'input': {'$toDate': '$_id'},
								'to': datetime
							}
						}
					}
				}
			)

		# Unwind, i.e converts
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'flag': 8740,
		#  'run': 714,
		#  'channel': 3,
		#  'msg': ['msg1', 'msg2'],
		# Into:
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'flag': 8740,
		#  'run': 714,
		#  'channel': 3,
		#  'msg': 'msg1'},
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'flag': 8740,
		#  'run': 714,
		#  'channel': 3,
		#  'msg': 'msg2'},
		stages.append(
			{
				'$unwind': {
					'path': '$msg',
					'preserveNullAndEmptyArrays': True
				}
			}
		)

		if decompactify:

			# Converts
			#  {'_id': ObjectId('5c80d71154048002ca372208'),
			#  'stock': 1810112413252531,
			#  'alert': 697252385515010002,
			#  'flag': 8740,
			#  'run': 714,
			#  'msg': {'channel': 3, 'txt': 'test2'},
			#  'date': datetime.datetime(2019, 3, 7, 8, 32, 17)}
			# Into:
			#  {'_id': ObjectId('5c80d71154048002ca372208'),
			#  'stock': 1810112413252531,
			#  'alert': 697252385515010002,
			#  'flag': 8740,
			#  'run': 714,
			#  'msg': 'test2',
			#  'date': datetime.datetime(2019, 3, 7, 8, 32, 17),
			#  'channel': 3}
			#######################################################

			# Part 1
			# a) Copy channel info embedded in compact msg to log entry root level
			# b) If msg.txt exists, copy it to root level as 'msg'
			stages.append(
				{
					"$addFields": {
						"channel": {
							"$ifNull": ["$channel", "$msg.channel"]
						},
						"msg": {
							"$ifNull": [
								"$msg.txt",
								{"$ifNull": ["$msg", "$REMOVE"]}
							]
						},
					}
				}
			)

			# Part 2:
			# remove compact dict leftover
			# ('msg': {'channel': 3, 'txt': 'test2'})
			stages.append(
				{
					"$addFields": {
						"msg": {
							"$cond": {
								"if": {"$not": "$msg.channel"},
								"then": "$msg",
								"else": "$REMOVE"
							}
						}
					}
				}
			)

			# re-apply channel filter as $unwind may have included
			# log entries associated with other channel
			if hasattr(matcher, 'channel'):
				stages.append(
					{
						'$match': {
							'channel': matcher.channels
						}
					}
				)

		if remove:

			# Projection (last aggregation stage)
			proj = {'$project': {}}

			for el in remove:
				proj['$project'][el] = 0

			stages.append(proj)

		if verbose:
			print("Using following aggregation stages:")
			print(stages)

		log_entries = tuple(
			el for el in AmpelDB.get_collection("logs").aggregate(stages)
		)

		if simplify:
			for el in log_entries:
				print("%s %s" % (el['_id'], el['msg']))
			return

		if hexify and resolve_flag:

			for el in log_entries:
				el['flag'] = LogFlag(el['flag'])
				if 'comp' in el:
					el['comp'] = el['comp'].hex()

		else:

			if hexify:
				for el in log_entries:
					if 'comp' in el:
						el['comp'] = el['comp'].hex()

			if resolve_flag:
				for el in log_entries:
					el['flag'] = LogFlag(el['flag'])

		if self.read_only:
			return tuple(ReadOnlyDict(el) for el in log_entries)

		return log_entries



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/complement/T3ExtJournalAppender.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.06.2020
# Last Modified Date: 17.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from pymongo import MongoClient
from typing import Iterable, Optional, Union, List
from ampel.type import StockId
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.aux.filter.SimpleDictArrayFilter import SimpleDictArrayFilter
from ampel.content.JournalRecord import JournalRecord
from ampel.core.AmpelContext import AmpelContext
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.t3.complement.AbsT3DataAppender import AbsT3DataAppender
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.FlatAnyOf import FlatAnyOf
#from ampel.abstract.AbsApplicable import AbsApplicable


class T3ExtJournalAppender(AbsT3DataAppender):
	"""
	Import journal entries from a 'foreign' database, e.g. one created
	by a previous version of Ampel.
	"""

	mongo_resource: str = "resource.ext_mongo"
	db_name: str = "Ampel_data"
	sort: bool = True
	reverse: bool = True
	filter_config: Optional[
		Union[FilterCriterion, FlatAnyOf[FilterCriterion], AllOf[FilterCriterion]]
	] = None


	def __init__(self, context: AmpelContext, **kwargs) -> None:

		AmpelBaseModel.__init__(self, **kwargs)

		if self.filter_config:
			self.journal_filter: SimpleDictArrayFilter[JournalRecord] = SimpleDictArrayFilter(filters=self.filter_config)

		self.col = MongoClient(context.config.get(f'resource.{self.mongo_resource}')) \
			.get_database(self.db_name)\
			.get_collection("stock")


	def get_ext_journal(self, stock_id: StockId) -> Optional[List[JournalRecord]]:

		if ext_stock := next(self.col.find({'_id': stock_id}), None):
			if self.journal_filter:
				return self.journal_filter.apply(ext_stock['journal'])
			return ext_stock['journal']
		return None


	def complement(self, it: Iterable[AmpelBuffer]) -> None:

		for albuf in it:

			if 'stock' in albuf and isinstance(albuf['stock'], dict):

				if entries := self.get_ext_journal(albuf['stock']['_id']):

					entries.extend(albuf['stock']['journal'])

					if self.sort:
						dict.__setitem__(
							albuf['stock'], 'journal', sorted( # type: ignore[index]
								entries, key=lambda x: x['ts'], reverse=self.reverse
							)
						)
			else:
				print(albuf)
				raise ValueError("No stock information available")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/ingest/AbsStateT2Ingester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 10.03.2020
# Last Modified Date: 30.04.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Tuple, Union
from ampel.type import StockId, ChannelId
from ampel.base import abstractmethod
from ampel.abstract.ingest.AbsT2Ingester import AbsT2Ingester
from ampel.ingest.CompoundBluePrint import CompoundBluePrint


class AbsStateT2Ingester(AbsT2Ingester, abstract=True):


	@abstractmethod
	def ingest(self,
		stock_id: StockId,
		comp_blueprint: CompoundBluePrint,
		chan_selection: List[Tuple[ChannelId, Union[bool, int]]]
	) -> None:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/T3ChannelUnitRunner.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 21.06.2020
# Last Modified Date: 21.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Sequence
from ampel.type import ChannelId
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.core.AmpelContext import AmpelContext
from ampel.model.UnitModel import UnitModel
from ampel.t3.run.AbsT3UnitRunner import AbsT3UnitRunner
from ampel.t3.run.T3UnitRunner import T3UnitRunner


class T3ChannelUnitRunner(AbsT3UnitRunner):
	"""
	UnitRunner class that configures the underlying T3UnitRunner instance
	to filter and project AmpelBuffer instances with respect to a single channel.
	Essentially a shortcut class since the same functionality
	can be obtained using a properly configured T3UnitRunner instance.
	T3UnitRunner is not inherited because of the differences in model definitions.
	"""

	channel: ChannelId
	execute: Sequence[UnitModel]


	def __init__(self, context: AmpelContext, **kwargs) -> None:
		"""
		:param channel: channel name/id
		:param execute: sequence of T3 units to be executed
		"""

		self._unit_runner = T3UnitRunner(
			context = context,
			directives=[
				T3UnitRunner.RunDirective(
					filter = UnitModel(
						unit="T3AmpelBufferFilter",
						config={"channel": kwargs['channel']}
					),
					project = UnitModel(
						unit = "T3ChannelProjector",
						config = {"channel": kwargs['channel']}
					),
					execute = kwargs.pop('execute')
				)
			],
			**kwargs
		)


	def run(self, data: Sequence[AmpelBuffer]) -> None:
		self._unit_runner.run(data)

	def done(self) -> None:
		self._unit_runner.done()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/ChannelModel.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.10.2019
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, List, Optional, Dict, Any
from ampel.model.StrictModel import StrictModel
from ampel.model.purge.PurgeModel import PurgeModel


class ChannelModel(StrictModel):

	channel: Union[int, str]
	purge: PurgeModel = {
		'content': {'delay': 100, 'format': 'json', 'unify': True},
		'logs': {'delay': 50, 'format': 'csv'}
	} # type: ignore[assignment]
	active: bool = True
	hash: Optional[int]
	distrib: Optional[str]
	source: Optional[str]
	contact: Optional[str]
	access: Optional[List[str]]
	policy: List[str] = []


	def dict(self, **kwargs) -> Dict[str, Any]:
		if 'exclude_defaults' not in kwargs:
			kwargs['exclude_defaults'] = False
		return super().dict(**kwargs)



import pickle
from pathlib import Path

import pytest

from ampel.content.StockDocument import StockDocument
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.log.AmpelLogger import AmpelLogger, DEBUG
from ampel.t3.run.project.T3ChannelProjector import T3ChannelProjector


@pytest.fixture
def stock_doc() -> StockDocument:
    with open(Path(__file__).parent / "test-data" / "ZTF20abxvcrk.pkl", "rb") as f:
        return pickle.load(f)


@pytest.fixture
def logger():
    return AmpelLogger.get_logger(console={"level": DEBUG})


def strip_channel(jentries):
    return [{k: v for k, v in jentry.items() if k != "channel"} for jentry in jentries]


def test_single_channel(stock_doc, logger):
    target = "TDE_RANKING"
    before = AmpelBuffer(id=stock_doc["_id"], stock=stock_doc)
    all_channels = set(before["stock"]["channel"])
    assert target in all_channels, "target is in test case channel set"
    assert len(all_channels) > 1, "test case contains channels not in target"

    proj = T3ChannelProjector(channel=target, logger=logger)
    after = proj.project([before])[0]
    for field in "tag", "name", "channel":
        assert not isinstance(after["stock"][field], str), f"stock.{field} must be a set"
    assert after["stock"]["channel"] == [target]
    assert after["stock"]["modified"] == {target: before["stock"]["modified"][target]}

    before_no_channel = strip_channel(before["stock"]["journal"])
    after_no_channel = strip_channel(after["stock"]["journal"])

    # all entries have only the selected channel
    for jentry in after["stock"]["journal"]:
        if "channel" in jentry:
            assert jentry["channel"] == target

    for jentry, stripped in zip(before["stock"]["journal"], before_no_channel):
        if "channel" in jentry:
            try:
                after_no_channel.index(
                    stripped
                ), "entry with target channel was preserved"
            except ValueError:
                if target in jentry["channel"]:
                    raise


@pytest.mark.parametrize("logic_op", ["all_of", "any_of", "one_of"])
def test_multi_channel(stock_doc, logger, logic_op):
    target = {"RCF_2020B", "ZTF_SLSN"}
    before = AmpelBuffer(id=stock_doc["_id"], stock=stock_doc)
    all_channels = set(before["stock"]["channel"])
    assert target.intersection(all_channels), "target is in test case channel set"
    assert len(all_channels) > len(target), "test case contains channels not in target"

    proj = T3ChannelProjector(channel={logic_op: list(target)}, logger=logger)
    after = proj.project([before])[0]
    for field in "tag", "name", "channel":
        assert not isinstance(after["stock"][field], str), f"stock.{field} must be a set"
    assert set(after["stock"]["channel"]) == target
    assert after["stock"]["modified"].keys() == target

    before_no_channel = strip_channel(before["stock"]["journal"])
    after_no_channel = strip_channel(after["stock"]["journal"])

    # all entries have only the selected channel
    for jentry in after["stock"]["journal"]:
        if (channel := jentry.get("channel")) is None:
            continue
        if isinstance(channel, (int, str)):
            assert channel in target
        else:
            assert len(channel) > 1
            assert set(channel).issubset(target)

    for jentry, stripped in zip(before["stock"]["journal"], before_no_channel):
        if (channel := jentry.get("channel")) is None:
            continue
        try:
            after_no_channel.index(
                stripped
            ), "entry with channel in target was preserved"
        except ValueError:
            if (isinstance(channel, (int, str)) and channel in target) or set(
                channel
            ).issubset(target):
                raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/t1.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.08.2020
# Last Modified Date: 15.08.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson.int64 import Int64
from typing import Any, List, Union, Optional, Set, Dict, Sequence, Callable, List
from ampel.type import ChannelId, StockId, DataPointId
from ampel.content.Compound import Compound
from ampel.log.AmpelLogger import AmpelLogger


def get_datapoint_ids(compound: Compound, logger: Optional[AmpelLogger] = None) -> List[DataPointId]:

	dps_ids: List[DataPointId] = []

	# Load each datapoint referenced by the loaded compound
	for el in compound['body']:

		# Dict means custom policy/exclusion is set for this datapoint
		if isinstance(el, dict):
			if 'excl' in el:
				if logger:
					logger.debug("Ignoring excluded datapoint", extra={'id': el['id']})
				continue
			dpid = el['id']
		elif isinstance(el, (int, Int64)):
			dpid = el
		else:
			raise StopIteration('Invalid compound doc')

		dps_ids.append(dpid)

	return dps_ids



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/model/t3/LoaderDirective.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 09.12.2019
# Last Modified Date: 18.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Optional, Type, Literal
from ampel.content.StockDocument import StockDocument
from ampel.content.DataPoint import DataPoint
from ampel.content.Compound import Compound
from ampel.content.T2Document import T2Document
from ampel.content.LogDocument import LogDocument
from ampel.model.t3.AliasableModel import AliasableModel


models = {
	"stock": StockDocument,
	"t0": DataPoint,
	"t1": Compound,
	"t2": T2Document,
	"log": LogDocument
}

class LoaderDirective(AliasableModel):
	"""Specification of documents to load"""

	#: Source collection
	col: Literal["stock", "t0", "t1", "t2", "log"]
	model: Optional[Type] # TypedDict
	#: Mongo match expression to include in the query
	query_complement: Optional[Dict[str, Any]]
	# key "link_config" used in DBContentLoader
	options: Optional[Dict[str, Any]]

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		if not self.model and self.col in models:
			self.model = models[self.col]
		elif self.model and not hasattr(self.model, "__annotations__"):
			raise ValueError("TypedDict expected for parameter 'model'")

	def dict(self, **kwargs):
		# do not emit model if it was equivalent to the default
		rep = super().dict(**kwargs)
		if self.model == models[self.col]:
			del rep["model"]
		return rep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/db/DBContentLoader.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 13.01.2018
# Last Modified Date: 13.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from bson.codec_options import CodecOptions
from typing import Iterable, Union, Dict, Iterator, Optional, Literal

from ampel.type import StockId, ChannelId, StrictIterable, Tag
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.t3.LoaderDirective import LoaderDirective
from ampel.db.FrozenValuesDict import FrozenValuesDict
from ampel.log.utils import safe_query_dict
from ampel.log.AmpelLogger import AmpelLogger
from ampel.db.query.general import build_general_query
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.util.collections import ampel_iter, to_set
from ampel.core.AdminUnit import AdminUnit
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry

freeze_codec_options = CodecOptions(document_class=FrozenValuesDict)

# Monitoring counters
stat_db_loads = AmpelMetricsRegistry.counter(
	"loads",
	"Number of documents loaded",
	subsystem="db",
	labelnames=("col",)
)

class DBContentLoader(AdminUnit):

	logger: Optional[AmpelLogger]

	def load(self,
		stock_ids: Union[StockId, Iterator[StockId], StrictIterable[StockId]],
		directives: Iterable[LoaderDirective],
		channel: Union[None, ChannelId, AllOf[ChannelId], AnyOf[ChannelId], OneOf[ChannelId]] = None,
		tag: Optional[Dict[Literal['with', 'without'], Union[Tag, Dict, AllOf[Tag], AnyOf[Tag], OneOf[Tag]]]] = None,
		auto_project: bool = True,
		codec_options: CodecOptions = freeze_codec_options,
	) -> Iterable[AmpelBuffer]:
		"""
		:param tag: If specified, query selection critera will apply to all directives.
		:param channel: If specified, query selection critera will apply to all directives (except t0).
		None means all channels are considered (no criterium).

		:param directives: see LoaderDirective docstrings for more information.  Notes:
		- T3 unit configurations can use aliased directives. For example, the alias COMPOUND
		is translated into the directive LoaderDirective(col="t1")
		- LoaderDirective can contain "query_complement", which can be used among other thing to further
		sub-select t2 documents or select given states (such as the latest state)

		:param auto_project: each LoaderDirective is associated with a db collection with is itself associated with
		a default "content" typed dict. By default, we request a projection that projects only the fields defined
		in those TypedDict. Custom/admin fields would thus not be retrieved.
		Set this setting to False if it is not the whished behavior.
		"""

		logger = self.logger if self.logger else AmpelLogger.get_logger()

		if not directives:
			raise ValueError("Parameter 'directives' cannot be empty")

		col_set = {directive.col for directive in directives}

		# Note: the following operation will consume
		# stock_ids if it is an Iterator/Cursor
		register: Dict[StockId, AmpelBuffer] = {
			stock_id: AmpelBuffer(
				id = stock_id,
				t0 = [] if "t0" in col_set else None,
				t1 = [] if "t1" in col_set else None,
				t2 = [] if "t2" in col_set else None,
				log = [] if "log" in col_set else None,
			) for stock_id in ampel_iter(stock_ids)
		}

		for directive in directives:

			query = build_general_query(
				stock=register.keys(), channel=channel, tag=tag
			)

			if directive.col == "stock":
				query['_id'] = query.pop("stock")
			elif directive.col == "t0" and channel:
				query.pop('channel')

			# query 'stock' parameter primes over query complements
			if directive.query_complement:
				query = {**directive.query_complement, **query}

			if logger.verbose > 1: # log query parameters
				logger.debug(
					None, extra={
						'col': directive.col,
						'query': safe_query_dict(query, dict_key=None)
					}
				)

			# Retrieve pymongo cursor
			col = self.context.db.get_collection(directive.col)

			if codec_options:
				col = col.database.get_collection(col.name, codec_options=codec_options)

			# Note: codec_options freezes structures in dicts with depth level > 1
			cursor = col.find(
				filter=query, projection={
					k: 1 for k in directive.model.__annotations__.keys()
				} if auto_project else None
			)

			doc_counter = stat_db_loads.labels(directive.col)

			if directive.col in ("t1", "log"):
				count = 0
				for count, res in enumerate(cursor, 1):
					register[res['stock']][directive.col].append(res) # type: ignore[union-attr]
				doc_counter.inc(count)

			elif directive.col == "stock":
				count = 0
				for count, res in enumerate(cursor, 1):
					register[res['_id']]['stock'] = res
				doc_counter.inc(count)

			# Datapoints are channel-less and can be associated with multiple stocks
			elif directive.col == "t0":

				count = 0
				for count, res in enumerate(cursor, 1):

					# Upper limits can be attached to multiple transients
					for sid in res['stock']:

						# Some of which might not match with our query
						if sid not in register:
							continue

						# Add datapoints to snapdata (no need to recursive_freeze
						# since dict elements with level > 1 are frozen due to codec_options
						register[sid]['t0'].append(res) # type: ignore[union-attr]
				doc_counter.inc(count)

			# Potentialy link config dict used by science records
			elif directive.col == "t2":

				# whether to link corresponding run config
				link_config = getattr(directive, 'options') and \
					directive.options.get('link_config', False) # type: ignore[union-attr]

				if link_config:
					config_keys = self.context.config.get('t2.config_keys', dict)
					if not config_keys:
						raise ValueError("Cannot load t2 configs")

				count = 0
				for count, res in enumerate(cursor, 1):

					if link_config:
						# link run_config dict
						res['config'] = res[config_keys[res['config']]] # type: ignore[index]

					register[res['stock']]['t2'].append(res) # type: ignore[union-attr]
				doc_counter.inc(count)

			else:
				raise ValueError(
					f"Unrecognized LoaderDirective: {directive.dict()}"
				)

		# Ensure that T2Documents are ordered by id
		for buf in register.values():
			if (buf_t2 := buf["t2"]) is not None:
				buf["t2"] = sorted(buf_t2, key=lambda doc: doc["_id"])

		if logger.verbose:
			s = f"Unique ids: {len(register)}"
			for col in (col_set - set(["stock"])):
				s += f", {col}: "
				s += str(sum([1 for k in register for el in register[k][col]])) # type: ignore
			logger.info(s)

		return register.values()


	@staticmethod
	def import_journal(tran_data, journal_entries, channels_set, logger):
		"""
		:param tran_data: instance of AmpelBuffer
		:type tran_data: :py:class:`AmpelBuffer <ampel.core.AmpelBuffer>`
		:param list(Dict) journal_entries:
		:param channels_set:
		:type channels_set: set(str), str
		"""

		# Save journal entries related to provided channels
		for entry in journal_entries:

			# Not sure if those entries will exist. Time will tell.
			if entry.get('channel') is None:
				logger.warn(
					'Ignoring following channel-less journal entry: %s' % str(entry),
					extra={'stock': tran_data.tran_id}
				)
				continue

			if channels_set == "Any":
				chans_intersec = "Any"
			else:
				# Set intersection between registered and requested channels (if any)
				chans_intersec = (
					entry['channel'] if channels_set is None
					else (channels_set & to_set(entry['channels']))
				)

			# Removing embedded 'channels' key/value and add journal entry
			# to transient data while maintaining the channels association
			tran_data.add_journal_entry(
				chans_intersec,
				# journal entry without the channels key/value
				{k: v for k, v in entry.items() if k != 'channel'}
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/config/LogicSchemaIterator.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 12.10.2018
# Last Modified Date: 04.02.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf


class LogicSchemaIterator:
	"""
	.. sourcecode:: python\n

		for schema in (a,b,c):
			print("Schema: %s" % schema)
			for i, el in enumerate(LogicSchemaIterator(schema)):
				print("Iteration %i: %s" % (i+1, el))

		Schema: 'a'
		Iteration 1: a

		Schema: {'any_of': ['a', 'b', 'c']}
		Iteration 1: a
		Iteration 2: b
		Iteration 3: c

		Schema: {'all_of': ['a', 'b', 'c']}
		Iteration 1: ['a', 'b', 'c']

		Schema: {'any_of': [{'all_of': ['a', 'b']}, {'all_of': ['a', 'c']}, 'd']}
		Iteration 1: ['a', 'b']
		Iteration 2: ['a', 'c']
		Iteration 3: d

		'with' scenario:
			- on scalars: if matches, then pass
			- on lists: if all matches, then pass
		'without' scenario:
			- on scalars: if matches, then discards
			- on lists: if any matches, then discards
	"""

	def __init__(self, arg, in_type=(str, int)):
		""" """
		self.arg = arg

		if isinstance(arg, (AllOf, AnyOf, OneOf)):
			arg = arg.dict()

		if isinstance(arg, in_type):
			self.values = [arg]

		elif isinstance(arg, dict):
			if "any_of" in arg:
				self.values = arg['any_of']
			elif 'all_of' in arg or 'one_of' in arg:
				self.values = [arg]
			else:
				raise ValueError("Unsupported format")
		else:
			raise ValueError("Unsupported type: %s" % type(arg))

		self.count = -1


	def __iter__(self):
		""" """
		return self


	def __next__(self):
		""" """
		try:
			self.count += 1
			if isinstance(self.values[self.count], dict):
				if 'all_of' in self.values[self.count]:
					return self.values[self.count]['all_of']
				if 'one_of' in self.values[self.count]:
					return self.values[self.count]['one_of']
			return self.values[self.count]
		except IndexError:
			raise StopIteration



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/run/T3UnitRunner.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 06.01.2020
# Last Modified Date: 10.06.2020
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from dataclasses import dataclass, field
from typing import Dict, Optional, Sequence, List, Type, Tuple, TypeVar, get_args, Union
from pydantic import validator

from ampel.log import VERBOSE, AmpelLogger, LogFlag
from ampel.log.utils import report_exception
from ampel.log.handlers.ChanRecordBufHandler import ChanRecordBufHandler
from ampel.log.handlers.DefaultRecordBufferingHandler import DefaultRecordBufferingHandler
from ampel.core.AmpelBuffer import AmpelBuffer
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.core.JournalUpdater import JournalUpdater
from ampel.view.SnapView import SnapView
from ampel.util.mappings import build_unsafe_dict_id
from ampel.util.freeze import recursive_freeze
from ampel.model.StrictModel import StrictModel
from ampel.model.UnitModel import UnitModel
from ampel.struct.JournalTweak import JournalTweak
from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.t3.run.AbsT3UnitRunner import AbsT3UnitRunner
from ampel.t3.run.filter.AbsT3Filter import AbsT3Filter
from ampel.t3.run.project.AbsT3Projector import AbsT3Projector


@dataclass
class RunBlock:
	"""
	data class used internally by T3UnitRunner
	"""
	filter: Optional[AbsT3Filter] = None
	projector: Optional[AbsT3Projector] = None
	units: List[Tuple[AbsT3Unit, Type]] = field(default_factory=list)


class T3UnitRunner(AbsT3UnitRunner):
	"""
	Default implementation handling the instructions defined
	by :class:`T3Directive.run <ampel.model.t3.T3Directive.T3Directive>`.
	"""

	class RunDirective(StrictModel):
		"""
		Internal model used for field 'directives' of T3UnitRunner
		"""

		#: unit to use for down-selection of stocks
		filter: Optional[UnitModel]
		#: unit to use for projection of each stock
		project: Optional[UnitModel]
		#: units to use to execute T3s
		execute: Sequence[UnitModel]

		@validator('execute', pre=True)
		def cast_to_sequence_if_need_be(cls, v):
			if isinstance(v, dict):
				return [v]
			return v

	#: Processing specification
	directives: Sequence[RunDirective]


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.run_blocks: List[RunBlock] = []
		self.units: Dict[int, Tuple[AbsT3Unit, Type[SnapView]]] = {}
		debug = self.logger.verbose > 1

		if debug:
			self.logger.debug("Setting up unit runner")

		for directive in self.directives:

			rb = RunBlock()

			if directive.filter:

				if debug:
					self.logger.debug(f"Setting up filter {directive.filter.unit_name}")

				rb.filter = self.context.loader.new(
					unit_model = directive.filter,
					unit_type = AbsT3Filter,
					logger = self.logger
				)

			if directive.project:

				if debug:
					self.logger.debug(f"Setting up projector {directive.project.unit_name}")

				rb.projector = AuxUnitRegister.new_unit(
					unit_model = directive.project,
					sub_type = AbsT3Projector,
					logger = self.logger
				)

			if debug:
				self.logger.debug(f"Creating buffering logging handler ({self.channel or 'no channel'})")

			self.buf_hdlr: Union[ChanRecordBufHandler, DefaultRecordBufferingHandler] = \
				ChanRecordBufHandler(self.logger.level, self.channel) if self.channel \
				else DefaultRecordBufferingHandler(self.logger.level)

			for exec_def in directive.execute:

				# Hash config associated with this T3 unit
				config_id = build_unsafe_dict_id(
					{
						"unit": exec_def.unit_name,
						"config": self.context.loader.get_init_config(
							exec_def.unit_name,
							exec_def.config,
							exec_def.override,
							resolve_secrets = False,
						)
					},
					ret = int
				)

				# If a T3 instance exists with the same config, re-use it
				if config_id in self.units:
					rb.units.append(self.units[config_id])
					continue

				if self.logger.verbose:
					if exec_def.config:
						self.logger.log(VERBOSE,
							f"Instantiating unit {exec_def.unit_name} (id: ..{str(config_id)[-6:]})"
						)
					else:
						self.logger.log(VERBOSE, f"Instantiating unit {exec_def.unit_name}")

				# Spawn unit instance
				unit_instance = self.context.loader.new_base_unit(
					unit_model=exec_def,
					logger = AmpelLogger.get_logger(
						base_flag = (getattr(self.logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
						console = False,
						handlers = [self.buf_hdlr]
					),
					sub_type=AbsT3Unit,
					context=self.run_context
				)

				# Get view type from class definition (Generic[T])
				View: Type = get_args(unit_instance.__class__.__orig_bases__[0])[0] # type: ignore
				if type(View) is TypeVar:
					View = SnapView

				# Save unit instance in global unit register
				self.units[config_id] = unit_instance, View

				rb.units.append(self.units[config_id])

			self.run_blocks.append(rb)


	def run(self, data: Sequence[AmpelBuffer]) -> None:

		try:

			jupdater = JournalUpdater(
				ampel_db = self.context.db, tier = 3, run_id = self.run_id,
				process_name = self.process_name, logger = self.logger,
				raise_exc = self.raise_exc, extra_tag = self.extra_journal_tag,
				update_journal = self.update_journal,
			)

			for i, run_block in enumerate(self.run_blocks):

				if self.logger.verbose > 1:
					self.logger.debug(f"Running run-block {i}")

				if run_block.filter:

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block filter")

					data = run_block.filter.filter(data)

				if run_block.projector:

					if self.logger.verbose:
						self.logger.log(VERBOSE, "Applying run-block projection")

					data = run_block.projector.project(data)

				for t3_unit, View in run_block.units:

					# python: we're all consenting adults
					# ampel: let's rather be suspicious consenting adults
					# We cast ampel buffers into views possibly redundantly (for each sub-unit)
					# since there is no real read-only struct in python

					if self.logger.verbose:
						self.logger.log(VERBOSE, f"Creating views for {t3_unit.__class__.__name__}")

					views = tuple(View(**recursive_freeze(el)) for el in data)

					if ret := t3_unit.add(views):

						unit_name = t3_unit.__class__.__name__

						if isinstance(ret, dict):
							for k, v in ret.items():
								jupdater.add_record(stock=k, jextra=v, unit=unit_name)

						elif isinstance(ret, JournalTweak):
							jupdater.add_record(
								stock = [sv.id for sv in views], jextra = ret, unit = unit_name
							)

						else:
							self.logger.error(
								f"Unsupported result type returned by {unit_name}: {type(ret)}"
							)

					else:
						jupdater.add_record(
							stock = [sv.id for sv in views],
							unit = t3_unit.__class__.__name__
						)

					if self.buf_hdlr.buffer:
						self.buf_hdlr.forward(self.logger)

				if self.update_journal:
					jupdater.flush()

		except Exception as e:

			if self.raise_exc:
				raise e

			# Try to insert doc into trouble collection (raises no exception)
			report_exception(
				self.context.db, self.logger, exc=e,
				info={'process': self.process_name},
			)

		finally:
			if self.buf_hdlr.buffer:
				self.buf_hdlr.forward(self.logger)

	def done(self) -> None:
		try:

			for run_block in self.run_blocks:
				for t3_unit, *_ in run_block.units:
					t3_unit.done()
					if self.buf_hdlr.buffer:
						self.buf_hdlr.forward(self.logger)

		except Exception as e:

			if self.raise_exc:
				raise e

			# Try to insert doc into trouble collection (raises no exception)
			report_exception(self.context.db, self.logger, exc=e)

		finally:
			if self.buf_hdlr.buffer:
				self.buf_hdlr.forward(self.logger)
